import os
import sys
import unittest
import numpy as np

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack import calc_eta, calc_fwhm, deriv_eta, deriv_fwhm
from TRXASprefitpack.mathfun.deriv_tst import test_num_deriv
from TRXASprefitpack import dmp_osc_conv_gau, dmp_osc_conv_cauchy
from TRXASprefitpack import deriv_dmp_osc_conv_gau, deriv_dmp_osc_conv_cauchy

class TestDerivDmpOscConvIRF(unittest.TestCase):
    def test_deriv_dmp_osc_conv_gau_1(self):
        '''
        Test gradient of convolution of damped oscillation and gaussian irf (tau: 1, fwhm: 0.15, period: 0.5, phase: pi/4)
        '''
        tau = 1
        fwhm = 0.15
        period = 0.5
        phase = np.pi/4
        t = np.linspace(-1, 100, 2001)
        tst = deriv_dmp_osc_conv_gau(t, fwhm, 1/tau, period, phase)
        ref = test_num_deriv(dmp_osc_conv_gau, t, fwhm, 1/tau, period, phase)
        result = np.isclose(tst, ref)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)

    def test_deriv_dmp_osc_conv_cauchy_1(self):
        '''
        Test gradient of convolution of damped oscillation and cauchy irf (tau: 1, fwhm: 0.15, period: 0.5, phase: pi/4)
        '''
        tau = 1
        fwhm = 0.15
        period = 0.5
        phase = np.pi/4
        t = np.linspace(-1, 100, 2001)
        tst = deriv_dmp_osc_conv_cauchy(t, fwhm, 1/tau, period, phase)
        ref = test_num_deriv(dmp_osc_conv_cauchy, t, fwhm, 1/tau, period, phase)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)

    def test_deriv_exp_conv_pvoigt(self):
        '''
        Test gradient of convolution of damped oscillation and pseudo voigt irf (tau: 1, fwhm_G: 0.1, fwhm_L: 0.15, period: 0.5, phase: pi/4)
         Note. not implemented in mathfun module, check implementation in res_grad_osc function
        '''

        def tmp_fun(t, fwhm_G, fwhm_L, k, period, phase):
            fwhm = calc_fwhm(fwhm_G, fwhm_L)
            eta = calc_eta(fwhm_G, fwhm_L)
            gau = dmp_osc_conv_gau(t, fwhm, k, period, phase)
            cauchy = dmp_osc_conv_cauchy(t, fwhm, k, period, phase)
            return gau + eta*(cauchy-gau)
        
        tau_1 = 1
        fwhm_G = 0.1
        fwhm_L = 0.15
        period = 0.5
        phase = np.pi/4
        t = np.linspace(-1, 100, 2001)

        grad = np.empty((t.size, 6))
        fwhm = calc_fwhm(fwhm_G, fwhm_L)
        eta = calc_eta(fwhm_G, fwhm_L)
        dfwhm_G, dfwhm_L = deriv_fwhm(fwhm_G, fwhm_L)
        deta_G, deta_L = deriv_eta(fwhm_G, fwhm_L)
        diff = dmp_osc_conv_cauchy(t, fwhm, 1/tau_1, period, phase) - \
            dmp_osc_conv_gau(t, fwhm, 1/tau_1, period, phase)
        grad_gau = deriv_dmp_osc_conv_gau(t, fwhm, 1/tau_1, period, phase)
        grad_cauchy = deriv_dmp_osc_conv_cauchy(t, fwhm, 1/tau_1, period, phase)
        grad_tot = grad_gau + eta*(grad_cauchy-grad_gau)
        grad[:, 0] = grad_tot[:, 0]; grad[:, 3] = grad_tot[:, 2]
        grad[:, 4] = grad_tot[:, 3]; grad[:, 5] = grad_tot[:, 4]
        grad[:, 1] = dfwhm_G*grad_tot[:, 1] + deta_G*diff
        grad[:, 2] = dfwhm_L*grad_tot[:, 1] + deta_L*diff

        ref = test_num_deriv(tmp_fun, t, fwhm_G, fwhm_L, 1/tau_1, period, phase)

        result = np.allclose(grad, ref)
        self.assertEqual(result, True)

if __name__ == '__main__':
    unittest.main()


# fit tscan
# fitting tscan data
# Using sum of exponential decay convolved with
# normalized gaussian distribution
# normalized cauchy distribution
# normalized pseudo voigt profile
# (Mixing parameter eta is fixed according to
#  Journal of Applied Crystallography. 33 (6): 1311–1316.)

from typing import Optional
import os
from pathlib import Path
import argparse
import numpy as np
import matplotlib.pyplot as plt
from ..driver import TransientResult, save_TransientResult
from ..driver import fit_transient_exp, fit_transient_dmp_osc, fit_transient_both
from .misc import read_data

plt.rcParams['figure.figsize'] = (10, 5)

FITDRIVER = {'decay': fit_transient_exp, 'osc': fit_transient_dmp_osc, 'both': fit_transient_both}

def save_TransientResult_txt(result: TransientResult, dirname: str):
      '''
      save fitting result to the text file

      Args:
       result: fitting result
       dirname: name of the directory in which text files for fitting result are saved.
       name_of_dset: name of each data sets. If `name_of_dset` is None then it is set to [1,2,3,....]
       t: sequence of scan range
       eps: sequence of estimated error of each datasets
      
      Returns:
       `fit_summary.txt`: Summary for the fitting result
       `weight_{name_of_dset[i]}.txt`: Weight of each model component of i th dataset
       `phase_{name_of_dset[i]}.txt`: Phase factor of each oscillation component of i th dataset
       `fit_{name_of_dset[i]}.txt`: fitting curve for i th dataset
       `fit_osc_{name_of_dset[i]}.txt`: oscillation part of fitting curve for i th dataset [model = 'both']
       `fit_decay_{name_of_dset[i]}.txt`: decay part of fitting curve for i th dataset [model = 'both']
       `res_{name_f_dset[i]}_j.txt`: residual (fit-data) curve for j th scan of i th data
                                     The format of text file is (t, res, eps)

      
      Note:
       If `dirname` directory is not exists, it creates `dirname` directory.
      '''
      if not (Path.cwd()/dirname).exists():
            os.mkdir(dirname)
      
      with open(f'{dirname}/fit_summary.txt', 'w') as f:
            f.write(str(result))
      
      for i in range(len(result['t'])):
            coeff_fmt = result['eps'][i].shape[1]*['%.8e']
            fit_fmt = (1+result['eps'][i].shape[1])*['%.8e']
            coeff_header_lst = []
            fit_header_lst = ['time_delay']
            for j in range(result['eps'][i].shape[1]):
                  res_save = np.vstack((result['t'][i], result['res'][i][:,j], result['eps'][i][:,j])).T
                  np.savetxt(f"{dirname}/res_{result['name_of_dset'][i]}_{j+1}.txt", res_save,
                  fmt=['%.8e', '%.8e', '%.8e'], 
                  header=f"time_delay \t res_{result['name_of_dset'][i]}_{j+1} \t eps")
                  fit_header_lst.append(f"fit_{result['name_of_dset'][i]}_{j+1}")
                  coeff_header_lst.append(f"tscan_{result['name_of_dset'][i]}_{j+1}")
            
            fit_header = '\t'.join(fit_header_lst)
            coeff_header = '\t'.join(coeff_header_lst)

            np.savetxt(f"{dirname}/weight_{result['name_of_dset'][i]}.txt", result['c'][i], fmt=coeff_fmt,
            header=coeff_header)
            if result['model'] in ['dmp_osc', 'both']:
                np.savetxt(f"{dirname}/phase_{result['name_of_dset'][i]}.txt", result['phase'][i], fmt=coeff_fmt,
                header=coeff_header)   
            fit_save = np.vstack((result['t'][i], result['fit'][i].T)).T
            np.savetxt(f"{dirname}/fit_{result['name_of_dset'][i]}.txt", fit_save, fmt=fit_fmt, header=fit_header)
            if result['model'] == 'both':
                  fit_decay_save = np.vstack((result['t'][i], result['fit_decay'][i].T)).T
                  np.savetxt(f"{dirname}/fit_decay_{result['name_of_dset'][i]}.txt", fit_decay_save, fmt=fit_fmt, header=fit_header)
                  fit_osc_save = np.vstack((result['t'][i], result['fit_osc'][i].T)).T
                  np.savetxt(f"{dirname}/fit_osc_{result['name_of_dset'][i]}.txt", fit_osc_save, fmt=fit_fmt, header=fit_header)

      return

def plot_TransientResult(result: TransientResult, save_fig: Optional[str] = None):
      '''
      plot fitting Result

      Args:
       result: fitting result
       save_fig: prefix of saved png plots. If `save_fig` is `None`, plots are displayed instead of being saved.
      '''
      
      start = 0
      for i in range(len(result['t'])):
            for j in range(result['intensity'][i].shape[1]):
                  fig = plt.figure(start+j+1)
                  title = f'{result["name_of_dset"][i]} scan #{j+1}'
                  subtitle = f"Chi squared: {result['red_chi2_ind'][i][j]: .2f}"
                  plt.suptitle(title)
                  sub1 = fig.add_subplot(221)
                  sub1.set_title(subtitle)
                  sub1.errorbar(result['t'][i], result['intensity'][i][:, j], result['eps'][i][:, j], marker='o', mfc='none',
                  label=f'expt {title}', linestyle='none', color='black')
                  sub1.plot(result['t'][i], result['fit'][i][:, j], label=f'fit {title}', color='red')
                  sub1.legend()
                  sub2 = fig.add_subplot(222)
                  sub2.set_title(subtitle)
                  sub2.errorbar(result['t'][i], result['intensity'][i][:, j], result['eps'][i][:, j], marker='o', mfc='none',
                  label=f'expt {title}', linestyle='none', color='black')
                  sub2.plot(result['t'][i], result['fit'][i][:, j], label=f'fit {title}', color='red')
                  sub2.legend()
                  sub2.set_xlim(-10*result['fwhm'], 20*result['fwhm'])
                  sub3 = fig.add_subplot(223)
                  if result['model'] in ['decay', 'osc']:
                        sub3.errorbar(result['t'][i], result['res'][i][:, j], 
                        result['eps'][i][:, j], marker='o', mfc='none', 
                        label=f'res {title}', linestyle='none', color='black')
                  else:
                        sub3.errorbar(result['t'][i], result['intensity'][i][:, j]-result['fit_decay'][i][:, j], 
                        result['eps'][i][:, j], marker='o', mfc='none', label=f'expt osc {title}', 
                        linestyle='none', color='black')
                        sub3.plot(result['t'][i], result['fit_osc'][i][:, j], 
                        label=f'fit osc {title}', color='red')
                  sub3.legend()

                  sub4 = fig.add_subplot(224)
                  if result['model'] in ['decay', 'osc']:
                        sub4.errorbar(result['t'][i], result['res'][i][:, j], 
                        result['eps'][i][:, j], marker='o', mfc='none', 
                        label=f'res {title}', linestyle='none', color='black')
                  else:
                        sub4.errorbar(result['t'][i], result['intensity'][i][:, j]-result['fit_decay'][i][:, j], 
                        result['eps'][i][:, j], marker='o', mfc='none', 
                        label=f'expt osc {title}', linestyle='none', color='black')
                        sub4.plot(result['t'][i], result['fit_osc'][i][:, j], 
                        label=f'fit osc {title}', color='red')
                  sub4.set_xlim(-10*result['fwhm'], 20*result['fwhm'])
                  sub4.legend()
                  if save_fig is not None:
                        plt.savefig(f'./{save_fig}/{result["name_of_dset"][i]}_{j+1}.png')
            start = start + result['intensity'][i].shape[1]
      if save_fig is None:
            plt.show()
      return


description = '''
fit tscan: fitting experimental time trace spectrum data with the convolution of the sum of 
1. exponential decay (mode = decay)
2. damped oscillation (mode = osc)
3. exponential decay, damped oscillation (mode=both)
and irf function
There are three types of irf function (gaussian, cauchy, pseudo voigt)
To calculate the contribution of each life time component, it solve least linear square problem via scipy linalg lstsq module.
'''

epilog = '''
*Note

1. The number of time zero parameter should be same as the
   total number of scan to fit.

2. Every scan file whose prefix of filename is same should have same scan range

3. if you set shape of irf to pseudo voigt (pv), then
   you should provide two full width at half maximum
   value for gaussian and cauchy parts, respectively.

4. If you did not set tau and `mode=decay` then `--no_base` option is discouraged.

5. If you set `mode=decay` then any parameter whoose subscript is `osc` is discarded (i.e. tau_osc, period_osc).

6. If you set `mode=osc` then `tau` parameter is discarded. Also, baseline feature is not included in fitting function.

7. The number of tau_osc and period_osc parameter should be same

8. If you set `mode=both` then you should set `tau`, `tau_osc` and `period_osc`. However the number of `tau` and `tau_osc` need not to be same.
'''

mode_help = '''
Mode of fitting
 `decay`: fitting with the sum of the convolution of exponential decay and instrumental response function
 `osc`: fitting with the sum of the convolution of damped oscillation and instrumental response function
 `both`: fitting with the sum of both decay and osc
'''

irf_help = '''
shape of instrument response functon
g: gaussian distribution
c: cauchy distribution
pv: pseudo voigt profile, linear combination of gaussian distribution and cauchy distribution 
    pv = eta*c+(1-eta)*g 
    the uniform fwhm parameter and 
    mixing parameter are determined according to Journal of Applied Crystallography. 33 (6): 1311–1316. 
'''

fwhm_G_help = '''
full width at half maximum for gaussian shape
It would not be used when you set cauchy irf function
'''

fwhm_L_help = '''
full width at half maximum for cauchy shape
It would not be used when you did not set irf or use gaussian irf function
'''

do_glb_help='''
Whether or not use Global optimization Method.
When do_glb option is set then `basinhopping` global optimization algorithm will be used.
'''


def fit_tscan():

    tmp = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=tmp,
                                     description=description,
                                     epilog=epilog)
    parser.add_argument('--mode', default='decay', choices=['decay', 'osc', 'both'],
    help=mode_help)
    parser.add_argument('--irf', default='g', choices=['g', 'c', 'pv'],
                        help=irf_help)
    parser.add_argument('--fwhm_G', type=float,
                        help=fwhm_G_help)
    parser.add_argument('--fwhm_L', type=float,
                        help=fwhm_L_help)
    parser.add_argument('prefix', nargs='+',
                        help='prefix for tscan files ' +
                        'It will read prefix_i.txt')
    parser.add_argument('--num_file', type=int, nargs='+',
                         help='number of scan file corresponding to each prefix')
    parser.add_argument('-t0', '--time_zeros', type=float, nargs='+',
                        help='time zeros for each tscan')
    parser.add_argument('-t0f', '--time_zeros_file',
                        help='filename for time zeros of each tscan')
    parser.add_argument('--tau', type=float, nargs='*',
                        help='lifetime of each decay component [mode: decay, both]')
    parser.add_argument('--tau_osc', type=float, nargs='+',
                        help='lifetime of each damped oscillation component [mode: osc, both]')
    parser.add_argument('--period_osc', type=float, nargs='+',
                        help='period of the vibration of each damped oscillation component [mode: osc, both]')
    parser.add_argument('--no_base', action='store_false',
                        help='exclude baseline for fitting [mode: decay, both]')
    parser.add_argument('--fix_irf', action='store_true',
    help='fix irf parameter (fwhm_G, fwhm_L) during fitting process')
    parser.add_argument('--fix_t0', action='store_true',
    help='fix time zero parameter during fitting process.')
    parser.add_argument('--do_glb', action='store_true',
    help=do_glb_help)
    parser.add_argument('-o', '--outdir', default='out',
                        help='name of directory to store output files')
    parser.add_argument('--save_fig', action='store_true',
    help='save plot instead of display')
    args = parser.parse_args()

    prefix = np.array(args.prefix, dtype=str)
    num_file = np.array(args.num_file, dtype=int)

    irf = args.irf
    if irf == 'g':
        if args.fwhm_G is None:
            raise Exception('You are using gaussian irf, so you should set fwhm_G!\n')
        else:
            fwhm_init = args.fwhm_G
    elif irf == 'c':
        if args.fwhm_L is None:
            raise Exception('You are using cauchy/lorenzian irf,' +
            'so you should set fwhm_L!\n')
        else:
            fwhm_init = args.fwhm_L
    else:
        if (args.fwhm_G is None) or (args.fwhm_L is None):
            raise Exception('You are using pseudo voigt irf,' +
            'so you should set both fwhm_G and fwhm_L!\n')
        else:
            fwhm_init = np.array([args.fwhm_G, args.fwhm_L])

    if (args.time_zeros is None) and (args.time_zeros_file is None):
        raise Exception('You should set either time_zeros or time_zeros_file!\n')

    elif args.time_zeros is None:
        t0_init = np.genfromtxt(args.time_zeros_file)
    else:
        t0_init = np.array(args.time_zeros)

    t = np.empty(prefix.size, dtype=object)
    intensity = np.empty(prefix.size, dtype=object)
    eps = np.empty(prefix.size, dtype=object)
    num_scan = np.sum(num_file)

    if num_scan != t0_init.size:
        raise Exception('Number of initial time zero parameter should be same as num_file parameter')

    for i in range(prefix.size):
        t[i] = np.genfromtxt(f'{prefix[i]}_1.txt')[:, 0]
        num_data_pts = t[i].size
        intensity[i], eps[i] = read_data(prefix[i], num_file[i], num_data_pts, 10)

    print(f'fitting with total {num_scan} data set!\n')

    bound_fwhm = None
    if args.fix_irf:
        if irf in ['g', 'c']:
            bound_fwhm = [(fwhm_init, fwhm_init)]
        else:
            bound_fwhm = [(fwhm_init[0], fwhm_init[0]), (fwhm_init[1], fwhm_init[1])]
    
    bound_t0 = None
    if args.fix_t0:
        bound_t0 = t0_init.size*[None]
        for i in range(t0_init.size):
            bound_t0[i] = (t0_init[i], t0_init[i])
    dargs = []
    base = False
    if args.mode in ['decay', 'both']:
        if args.tau is None:
            base = True
            tau_init = None
        else:
            tau_init = np.array(args.tau)
            base = args.no_base
        dargs.append(tau_init)
        if args.mode == 'decay':
            dargs.append(base)
    if args.mode in ['osc', 'both']:
        tau_osc_init = np.array(args.tau_osc)
        period_osc_init = np.array(args.period_osc)
        dargs.append(tau_osc_init)
        dargs.append(period_osc_init)
    
    if args.mode == 'both':
        dargs.append(base)

    result = FITDRIVER[args.mode](irf, fwhm_init, t0_init, *dargs, do_glb=args.do_glb, 
    bound_fwhm=bound_fwhm, bound_t0=bound_t0, name_of_dset=prefix, t=t, intensity=intensity, eps=eps)

    save_TransientResult(result, args.outdir)
    save_TransientResult_txt(result, args.outdir)
    print(result)
    if args.save_fig:
        plot_TransientResult(result, args.outdir)
    else:
        plot_TransientResult(result)

    return



'''
driver:
subpackage for driver routine of TRXASprefitpack
convolution of sum of exponential decay and instrumental response function 

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from .ads import sads, dads
from .transient_result import TransientResult, save_TransientResult, load_TransientResult
from .static_result import StaticResult, save_StaticResult, load_StaticResult
from .static_result import static_spectrum
from ._transient_exp import fit_transient_exp
from ._transient_dmp_osc import fit_transient_dmp_osc
from ._transient_both import fit_transient_both
from ._static_voigt import fit_static_voigt
from ._static_thy import fit_static_thy
from .anal_fit import CIResult, is_better_fit, confidence_interval

__all__ = ['sads', 'dads',
           'TransientResult', 'save_TransientResult', 'load_TransientResult',
           'StaticResult', 'save_StaticResult', 'load_StaticResult',
           'static_spectrum',
           'fit_transient_exp', 'fit_transient_dmp_osc', 'fit_transient_both',
           'fit_static_voigt', 'fit_static_thy'
           ,'CIResult', 'is_better_fit', 'confidence_interval']


'''
deriv_tst:
submodule to test derivative routine of mathfun subpackage

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

import numpy as np

def test_num_deriv(func: callable, *args, eps_rel=1e-8, eps_abs=5e-6):
    '''
    Test implementation of derivative via finite difference
    i th column: derivative of f w.r.t ith argument
    '''
    n = len(args)
    if isinstance(args[0], np.ndarray):
        num_grad = np.empty((args[0].size, n))
        for i in range(n):
            f_args = list(args)
            b_args = list(args)
            f_args[i] = f_args[i]*(1+eps_rel)+eps_abs
            b_args[i] = b_args[i]*(1-eps_rel)-eps_abs
            f_args = tuple(f_args)
            b_args = tuple(b_args)
            dx = f_args[i] - b_args[i]
            num_grad[:, i] = (func(*f_args)-func(*b_args))/dx
    else:
        num_grad = np.empty(n)
        for i in range(n):
            f_args = list(args)
            b_args = list(args)
            f_args[i] = f_args[i]*(1+eps_rel)+eps_abs
            b_args[i] = b_args[i]*(1-eps_rel)-eps_abs
            f_args = tuple(f_args)
            b_args = tuple(b_args)
            dx  = f_args[i] - b_args[i]
            num_grad[i] = (func(*f_args)-func(*b_args))/dx
    
    return num_grad


'''
peak_shape:
submodule for the mathematical functions for
peak shape function

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''
from typing import Union, Optional
import numpy as np
from scipy.special import erf, wofz

def edge_gaussian(e: Union[float, np.ndarray], fwhm_G: float) -> np.ndarray:
    '''
    Gaussian type edge function :math:(\\frac{1}{2}\\left(1+{erf}\\left(\\frac{x}{\\sigma\\sqrt{2}}\\right)\\right))

    Args:
     e: energy
     fwhm_G: full width at half maximum
    
    Returns:
     gaussian type edge shape function
    '''

    return (1+erf(2*np.sqrt(np.log(2))*e/fwhm_G))/2

def edge_lorenzian(e: Union[float, np.ndarray], fwhm_L: float) -> np.ndarray:
    '''
    Lorenzian type edge function :math:(0.5+\\frac{1}{\\pi}{arctan}\\left(\\frac{x}{\\gamma}\\right))

    Args:
     e: energy
     fwhm_L: full width at half maximum

    Returns:
     lorenzian type edge shape function
    '''

    return 0.5+np.arctan(2*e/fwhm_L)/np.pi

def voigt(e: Union[float, np.ndarray], fwhm_G: float, fwhm_L: float) -> Union[float, np.ndarray]:
    '''
    voigt: evaluates voigt profile function with full width at half maximum of gaussian part is fwhm_G and
    full width at half maximum of lorenzian part is fwhm_L

    Args:
     e: energy
     fwhm_G: full width at half maximum of gaussian part :math:`(2\\sqrt{2\\log(2)}\\sigma)`
     fwhm_L: full width at half maximum of lorenzian part :math:`(2\\gamma)`
    
    Returns:
     voigt profile 

    Note:
     * if fwhm_G is (<1e-8) it returns normalized lorenzian shape
     * if fwhm_L is (<1e-8) it returns normalized gaussian shape
    '''
    sigma = fwhm_G/(2*np.sqrt(2*np.log(2)))
    
    if fwhm_G < 1e-8:
        return fwhm_L/2/np.pi/(e**2+fwhm_L**2/4)

    if fwhm_L < 1e-8:
        return np.exp(-(e/sigma)**2/2)/(sigma*np.sqrt(2*np.pi))
    
    z = (e+complex(0,fwhm_L/2))/(sigma*np.sqrt(2))
    return wofz(z).real/(sigma*np.sqrt(2*np.pi))

def voigt_thy(e: np.ndarray, thy_peak: np.ndarray,
              fwhm_G: float, fwhm_L: float,
              peak_factor: Union[float, np.ndarray],
              policy: Optional[str] = 'shift') -> np.ndarray:

    '''
    Calculates normalized 
    voigt broadened theoretically calculated lineshape spectrum

    Args:
        e: energy 
        thy_peak: theoretical calculated peak position and intensity
        fwhm_G: full width at half maximum of gaussian shape 
        fwhm_L: full width at half maximum of lorenzian shape 
        peak_factor: Peak factor, its behavior depends on policy.
        policy ({'shift', 'scale', 'both'}): Policy to match discrepency 
         between experimental data and theoretical spectrum.

         * 'shift' : Default option, shift peak position by peak_factor
         * 'scale' : scale peak position by peak_factor
         * 'both' : both shift and scale peak postition. 

         `peak_factor` should be equal to the pair of `shift_factor` and `scale_factor`.

    Returns:
      normalized voigt broadened theoritical lineshape spectrum
    '''

    v_matrix = np.empty((e.size, thy_peak.shape[0]))
    peak_copy = np.copy(thy_peak[:, 0])
    if policy == 'shift':
      peak_copy = peak_copy + peak_factor
    elif policy == 'scale':
      peak_copy = peak_factor*peak_copy 
    else:
        peak_copy = peak_factor[1]*peak_copy + peak_factor[0]
    for i in range(peak_copy.size):
        v_matrix[:, i] = voigt(e-peak_copy[i], fwhm_G, fwhm_L)

    broadened_theory = v_matrix @ thy_peak[:, 1].reshape((peak_copy.size, 1))

    return broadened_theory.flatten()

def deriv_edge_gaussian(e: Union[float, np.ndarray], fwhm_G: float) -> np.ndarray:
    '''
    derivative of gaussian type edge

    Args:
     e: energy
     fwhm_G: full width at half maximum
    
    Returns:
     first derivative of gaussian edge function
    
    Note:

     * 1st column: df/de
     * 2nd column: df/d(fwhm_G)
    '''
    tmp = np.exp(-4*np.log(2)*(e/fwhm_G)**2)/np.sqrt(np.pi)

    grad_e = 2*np.sqrt(np.log(2))/fwhm_G*tmp
    grad_fwhm_G = -2*np.sqrt(np.log(2))*e/fwhm_G/fwhm_G*tmp

    if isinstance(e, np.ndarray):
        grad = np.empty((e.size, 2))
        grad[:, 0] = grad_e
        grad[:, 1] = grad_fwhm_G
    else:
        grad = np.empty(2)
        grad[0] = grad_e
        grad[1] = grad_fwhm_G

    return grad

def deriv_edge_lorenzian(e: Union[float, np.ndarray], fwhm_L: float) -> np.ndarray:
    '''
    derivative of lorenzian type edge

    Args:
     e: energy
     fwhm_G: full width at half maximum
    
    Returns:
     first derivative of lorenzian type function
    
    Note:

     * 1st column: df/de
     * 2nd column: df/d(fwhm_L)
    '''
    tmp = 1/np.pi/(e**2+fwhm_L**2/4)
    grad_e = fwhm_L*tmp/2
    grad_fwhm_L = -e*tmp/2


    if isinstance(e, np.ndarray):
        grad = np.empty((e.size, 2))
        grad[:, 0] = grad_e
        grad[:, 1] = grad_fwhm_L
    else:
        grad = np.empty(2)
        grad[0] = grad_e
        grad[1] = grad_fwhm_L

    return grad

def deriv_voigt(e: Union[float, np.ndarray], fwhm_G: float, fwhm_L: float) -> np.ndarray:
    '''
    deriv_voigt: derivative of voigt profile with respect to (e, fwhm_G, fwhm_L)

    Args:
     e: energy
     fwhm_G: full width at half maximum of gaussian part :math:(2\\sqrt{2\\log(2)}\\sigma)
     fwhm_L: full width at half maximum of lorenzian part :math:(2\\gamma)
    
    Returns:
     first derivative of voigt profile 

    Note:

     * 1st column: df/de
     * 2nd column: df/d(fwhm_G)
     * 3rd column: df/d(fwhm_L)

     if `fwhm_G` is (<1e-8) then,
    
     * 1st column: dl/de
     * 2nd column: 0
     * 3rd column: dL/d(fwhm_L)
    
     L means normalized lorenzian shape with full width at half maximum parameter: `fwhm_L`

     if `fwhm_L` is (<1e-8) it returns 

     * 1st column: dg/de
     * 2nd column: dg/d(fwhm_G)
     * 3rd column: 0
    
     g means normalized gaussian shape with full width at half maximum parameter: `fwhm_G`
    '''

    if fwhm_G < 1e-8:
        tmp = fwhm_L/2/np.pi/(e**2+fwhm_L**2/4)**2
        if isinstance(e, np.ndarray):
            grad = np.empty((e.size, 3))
            grad[:, 0] = - 2*e*tmp
            grad[:, 1] = 0
            grad[:, 2] = (1/np.pi/(e**2+fwhm_L**2/4)-fwhm_L*tmp)/2
        else:
            grad = np.empty(3)
            grad[0] = -2*e*tmp
            grad[1] = 0
            grad[2] = (1/np.pi/(e**2+fwhm_L**2/4)-fwhm_L*tmp)/2
        return grad
    
    sigma = fwhm_G/(2*np.sqrt(2*np.log(2)))
    if fwhm_L < 1e-8:
        tmp = np.exp(-(e/sigma)**2/2)/(sigma*np.sqrt(2*np.pi))
        if isinstance(e, np.ndarray):
            grad = np.empty((e.size, 3))
            grad[:, 0] = -e/sigma**2*tmp
            grad[:, 1] = ((e/sigma)**2-1)/fwhm_G*tmp
            grad[:, 2] = 0
        else:
            grad = np.empty(3)
            grad[0] = -e/sigma**2*tmp
            grad[1] = ((e/sigma)**2-1)/sigma*tmp
            grad[2] = 0
        return grad
    
    z = (e+complex(0,fwhm_L/2))/(sigma*np.sqrt(2))
    f = wofz(z)/(sigma*np.sqrt(2*np.pi))
    f_z = (complex(0, 2/np.sqrt(np.pi))-2*z*wofz(z))/(sigma*np.sqrt(2*np.pi))
    if isinstance(e, np.ndarray):
        grad = np.empty((e.size, 3))
        grad[:, 0] = f_z.real/(sigma*np.sqrt(2))
        grad[:, 1] = (-f/sigma-z/sigma*f_z).real/(2*np.sqrt(2*np.log(2)))
        grad[:, 2] = -f_z.imag/(2*np.sqrt(2)*sigma)
    else:
        grad = np.empty(3)
        grad[0] = f_z.real/(sigma*np.sqrt(2))
        grad[1] = (-f/sigma-z/sigma*f_z).real/(2*np.sqrt(2*np.log(2)))
        grad[2] = -f_z.imag/(2*np.sqrt(2)*sigma)
    return grad

def deriv_voigt_thy(e: np.ndarray, thy_peak: np.ndarray,
              fwhm_G: float, fwhm_L: float,
              peak_factor: Union[float, np.ndarray],
              policy: Optional[str] = 'shift') -> np.ndarray:

    '''
    Calculates derivative of normalized 
    voigt broadened theoretically calculated lineshape spectrum

    Args:
        e: energy 
        thy_peak: theoretical calculated peak position and intensity
        fwhm_G: full width at half maximum of gaussian shape 
        fwhm_L: full width at half maximum of lorenzian shape 
        peak_factor: Peak factor, its behavior depends on policy.
        policy ({'shift', 'scale', 'both'}): Policy to match discrepency 
         between experimental data and theoretical spectrum.

         * 'shift' : Default option, shift peak position by peak_factor
         * 'scale' : scale peak position by peak_factor
         * 'both' : both shift and scale peak postition. 

         `peak_factor` should be equal to the pair of `shift_factor` and `scale_factor`.

    Returns:
      derivative of normalized voigt broadened theoritical lineshape spectrum
    
    Note:
     * 1st column: df/d(fwhm_G)
     * 2nd column: df/d(fwhm_L)
     
     if policy is `shift` or `scale`,

     * 3rd column: df/d(peak_factor)

     if policy is `both`

     * 3rd column: df/d(peak_factor[0]), peak_factor[0]: shift factor
     * 4th column: df/d(peak_factor[1]), peak_factor[1]: scale factor
    '''

    deriv_voigt_tensor = np.empty((e.size, 3, thy_peak.shape[0]))
    peak_copy = np.copy(thy_peak[:, 0])
    if policy == 'shift':
      peak_copy = peak_copy + peak_factor
    elif policy == 'scale':
      peak_copy = peak_factor*peak_copy 
    else:
        peak_copy = peak_factor[1]*peak_copy + peak_factor[0]
    for i in range(peak_copy.size):
        deriv_voigt_tensor[:, :, i] = deriv_voigt(e-peak_copy[i], fwhm_G, fwhm_L)

    if policy in ['shift', 'scale']:
        grad = np.empty((e.size, 3))
    else:
        grad = np.empty((e.size, 4))
    
    grad[:, 0] = deriv_voigt_tensor[:, 1, :] @ thy_peak[:, 1]
    grad[:, 1] = deriv_voigt_tensor[:, 2, :] @ thy_peak[:, 1]

    if policy == 'shift':
        grad[:, 2] = -deriv_voigt_tensor[:, 0, :] @ thy_peak[:, 1]
    elif policy == 'scale':
        grad[:, 2] = -deriv_voigt_tensor[:, 0, :] @ \
        (thy_peak[:, 0]*thy_peak[:, 1])
    else:
        grad[:, 2] = -deriv_voigt_tensor[:, 0, :] @ thy_peak[:, 1]
        grad[:, 3] = -deriv_voigt_tensor[:, 0, :] @ \
            (thy_peak[:, 0]*thy_peak[:, 1])

    return grad



    


'''
res_decay:
submodule for residual function and gradient for fitting time delay scan with the
convolution of sum of exponential decay and instrumental response function 

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Optional, Sequence, Tuple
import numpy as np
from ..mathfun.irf import calc_eta, deriv_eta
from ..mathfun.irf import calc_fwhm, deriv_fwhm
from ..mathfun.A_matrix import make_A_matrix_gau, make_A_matrix_cauchy, fact_anal_A
from ..mathfun.exp_conv_irf import deriv_exp_sum_conv_gau, deriv_exp_sum_conv_cauchy

# residual and gradient function for exponential decay model 

def residual_decay(x0: np.ndarray, base: bool, irf: str, 
                   t: Optional[Sequence[np.ndarray]] = None, 
                   intensity: Optional[Sequence[np.ndarray]] = None, eps: Optional[Sequence[np.ndarray]] = None) -> np.ndarray:
    '''
    residual_decay
    scipy.optimize.least_squares compatible vector residual function for fitting multiple set of time delay scan with the
    sum of convolution of exponential decay and instrumental response function  

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}` to :math:`2+N_{scan}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}` to :math:`3+N_{scan}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Residual vector
    
    Note:
     each dataset does not include time range.
    '''
    x0 = np.atleast_1d(x0)
    
    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
    else:
            num_irf = 2
            fwhm = calc_fwhm(x0[0], x0[1])
            eta = calc_eta(x0[0], x0[1])

    num_t0 = 0; sum = 0
    for d in intensity:
        num_t0 = d.shape[1] + num_t0
        sum = sum + d.size
    
    chi = np.empty(sum)
    tau = x0[num_irf+num_t0:]
    if not base:
        k = 1/tau
    else:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau; k[-1] = 0

    end = 0; t0_idx = num_irf
    for ti,d,e in zip(t,intensity,eps):
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A = make_A_matrix_gau(ti-t0, fwhm, k)
            elif irf == 'c':
                A = make_A_matrix_cauchy(ti-t0, fwhm, k)
            else:
                A_gau = make_A_matrix_gau(ti-t0, fwhm, k)
                A_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
                A = A_gau + eta*(A_cauchy-A_gau)
            c = fact_anal_A(A, d[:,j], e[:,j])
            chi[end:end+d.shape[0]] = ((c@A) - d[:, j])/e[:, j]

            end = end + d.shape[0]
            t0_idx = t0_idx + 1

    return chi
    
def res_grad_decay(x0: np.ndarray, num_comp: int, base: bool, irf: str, 
                   fix_param_idx: Optional[np.ndarray] = None,
                   t: Optional[Sequence[np.ndarray]] = None, 
                   intensity: Optional[Sequence[np.ndarray]]= None, 
                   eps: Optional[Sequence[np.ndarray]]=None) -> Tuple[np.ndarray, np.ndarray]:
    '''
    res_grad_decay
    scipy.optimize.minimize compatible scalar residual and its gradient function for fitting multiple set of time delay scan with the
    sum of convolution of exponential decay and instrumental response function  

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}` to :math:`2+N_{scan}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':
        
        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}` to :math:`3+N_{scan}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     fix_param_idx: index for fixed parameter (masked array for `x0`)
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Tuple of scalar residual function :math:`(\\frac{1}{2}\\sum_i {res}^2_i)` and its gradient
    '''
    x0 = np.atleast_1d(x0)
    
    if irf in ['g', 'c']:
            num_irf = 1 
            fwhm = x0[0]
    else:
        num_irf = 2
        eta = calc_eta(x0[0], x0[1])
        fwhm = calc_fwhm(x0[0], x0[1])
        dfwhm_G, dfwhm_L = deriv_fwhm(x0[0], x0[1])
        deta_G, deta_L = deriv_eta(x0[0], x0[1])

    num_t0 = 0; sum = 0
    for d in intensity:
        num_t0 = num_t0 + d.shape[1]
        sum = sum + d.size

    tau = x0[num_irf+num_t0:]

    if not base:
        k = 1/tau
    else:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau; k[-1] = 0
    
    num_param = num_irf+num_t0+num_comp
    chi = np.empty(sum)
    df = np.empty((sum, tau.size+num_irf))
    grad = np.empty(num_param)

    end = 0; t0_idx = num_irf

    for ti,d,e in zip(t, intensity, eps):
        step = d.shape[0]
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A = make_A_matrix_gau(ti-t0, fwhm, k)
            elif irf == 'c':
                A = make_A_matrix_cauchy(ti-t0, fwhm, k)
            else:
                A_gau = make_A_matrix_gau(ti-t0, fwhm, k)
                A_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
                diff = A_cauchy-A_gau
                A = A_gau + eta*diff
            c = fact_anal_A(A, d[:,j], e[:,j])
            chi[end:end+step] = (c@A-d[:,j])/e[:, j]
                
            if irf == 'g':
                grad_tmp = deriv_exp_sum_conv_gau(ti-t0, fwhm, 1/tau, c, base)
            elif irf == 'c':
                grad_tmp = deriv_exp_sum_conv_cauchy(ti-t0, fwhm, 1/tau, c, base)
            else:
                grad_gau = deriv_exp_sum_conv_gau(ti-t0, fwhm, 1/tau, c, base)
                grad_cauchy = deriv_exp_sum_conv_cauchy(ti-t0, fwhm, 1/tau, c, base)
                grad_tmp = grad_gau + eta*(grad_cauchy-grad_gau)
   
            grad_tmp = np.einsum('i,ij->ij', 1/e[:,j], grad_tmp)
            if irf in ['g', 'c']:
                df[end:end+step, 0] = grad_tmp[:, 1]
            else:
                cdiff = (c@diff)/e[:, j]
                df[end:end+step, 0] = dfwhm_G*grad_tmp[:, 1]+deta_G*cdiff
                df[end:end+step, 1] = dfwhm_L*grad_tmp[:, 1]+deta_L*cdiff
            grad[t0_idx] = -chi[end:end+step]@grad_tmp[:, 0]
            df[end:end+step, num_irf:] = np.einsum('j,ij->ij', -1/tau**2, grad_tmp[:, 2:])
                
            end = end + step
            t0_idx = t0_idx + 1
    
    mask = np.ones(num_param, dtype=bool)
    mask[num_irf:num_irf+num_t0] = False
    grad[mask] = chi@df

    if fix_param_idx is not None:
        grad[fix_param_idx] = 0

    return np.sum(chi**2)/2, grad


'''
exp_conv_irf:
submodule for the mathematical functions for
irf (instrumental response function)

:copyright: 2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Union, Tuple
import numpy as np

def gau_irf(t: Union[float, np.ndarray], fwhm: float) -> Union[float, np.ndarray]:

    '''
    Compute gaussian shape irf function

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution

    Returns:
     normalized gaussian function.
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    return np.exp(-t**2/(2*sigma**2))/(sigma*np.sqrt(2*np.pi))

def cauchy_irf(t: Union[float, np.ndarray], fwhm: float) -> Union[float, np.ndarray]:

    '''
    Compute lorenzian shape irf function

    Args:
      t: time
      fwhm: full width at half maximum of cauchy distribution

    Returns:
     normalized lorenzian function.
    '''

    gamma = fwhm/2
    return gamma/np.pi/(t**2+gamma**2)

def calc_fwhm(fwhm_G: float, fwhm_L: float) -> float:
  '''
  Calculate uniform fwhm of both gaussian and cauchy component of 
  pseudo voigt profile with fwhm_G, fwhm_L based on 
  Journal of Applied Crystallography. 33 (6): 1311–1316.

  Args:
    fwhm_G: full width at half maximum of gaussian part
    fwhm_L: full width at half maximum of lorenzian part
  Returns:
   uniform fwhm parameter of both gaussian and cauchy component of
   pseudo voigt profile
  '''
  f = fwhm_G**5+2.69269*fwhm_G**4*fwhm_L + \
    2.42843*fwhm_G**3*fwhm_L**2 + \
    4.47163*fwhm_G**2*fwhm_L**3 + \
    0.07842*fwhm_G*fwhm_L**4 + \
    fwhm_L**5
  return f**(1/5)
  
def calc_eta(fwhm_G: float, fwhm_L: float) -> float:
  '''
  Calculate eta of pseudo voigt profile with fwhm_G, fwhm_L based on 
  Journal of Applied Crystallography. 33 (6): 1311–1316.

  Args:
    fwhm_G: full width at half maximum of gaussian part
    fwhm_L: full width at half maximum of lorenzian part
  Returns:
   maxing parameter eta
  '''
  f = fwhm_G**5+2.69269*fwhm_G**4*fwhm_L + \
    2.42843*fwhm_G**3*fwhm_L**2 + \
    4.47163*fwhm_G**2*fwhm_L**3 + \
    0.07842*fwhm_G*fwhm_L**4 + \
    fwhm_L**5
  f = f**(1/5)
  x = fwhm_L/f
  eta = 1.36603*x-0.47719*x**2+0.11116*x**3
  return eta

def pvoigt_irf(t: Union[float, np.ndarray], fwhm: float, eta: float) -> Union[float, np.ndarray]:
    '''
    Compute pseudo voight shape irf function
    (i.e. linear combination of gaussian and lorenzian function)

    Args:
      t: time
      fwhm: uniform full width at half paramter
      eta: mixing paramter
    Returns:
     pseudo voigt profile
    '''
    u = gau_irf(t, fwhm)
    v = cauchy_irf(t, fwhm)

    return u + eta*(v-u)

def deriv_fwhm(fwhm_G: float, fwhm_L: float) -> Tuple[float, float]:
  '''
  Computes gradient of uniform fwhm parameter of pseudo voigt approximation
  based on Journal of Applied Crystallography. 33 (6): 1311-1316.

  Args:
   fwhm_G: full width at half maximum of gaussian part
   fwhm_L: full width at half maximum of lorenzian part
  
  Returns:
   gradient of fwhm(fwhm_G, fwhm_L)
  '''
  f = fwhm_G**5+2.69269*fwhm_G**4*fwhm_L + \
    2.42843*fwhm_G**3*fwhm_L**2 + \
    4.47163*fwhm_G**2*fwhm_L**3 + \
    0.07842*fwhm_G*fwhm_L**4 + \
    fwhm_L**5
  df_fwhm_G = 5*fwhm_G**4+10.77076*fwhm_G**3*fwhm_L + \
    7.28529*fwhm_G**2*fwhm_L**2+8.94326*fwhm_G*fwhm_L**3 + \
      0.07842*fwhm_L**4
  df_fwhm_L = 5*fwhm_L**4 + 0.31368*fwhm_L**3*fwhm_G + \
    13.41489*fwhm_G**2*fwhm_L**2 + 4.85686*fwhm_L*fwhm_G**3 + \
      2.69269*fwhm_G**4
  return df_fwhm_G/f**(4/5)/5, df_fwhm_L/f**(4/5)/5



def deriv_eta(fwhm_G: float, fwhm_L: float) -> Tuple[float, float]:
  '''
  Calculate gradient of eta of pseudo voigt profile with fwhm_G, fwhm_L based on 
  Journal of Applied Crystallography. 33 (6): 1311–1316.

  Args:
    fwhm_G: full width at half maximum of gaussian part
    fwhm_L: full width at half maximum of lorenzian part
  Returns:
   gradient of eta(fwhm_G, fwhm_L)
  '''
  f = fwhm_G**5+2.69269*fwhm_G**4*fwhm_L + \
    2.42843*fwhm_G**3*fwhm_L**2 + \
    4.47163*fwhm_G**2*fwhm_L**3 + \
    0.07842*fwhm_G*fwhm_L**4 + \
    fwhm_L**5
  g = f**(-1/5); x = fwhm_L*g
  df_fwhm_G = 5*fwhm_G**4+10.77076*fwhm_G**3*fwhm_L + \
    7.28529*fwhm_G**2*fwhm_L**2+8.94326*fwhm_G*fwhm_L**3 + \
      0.07842*fwhm_L**4
  df_fwhm_L = 5*fwhm_L**4 + 0.31368*fwhm_L**3*fwhm_G + \
    13.41489*fwhm_G**2*fwhm_L**2 + 4.85686*fwhm_L*fwhm_G**3 + \
      2.69269*fwhm_G**4
  dx_fwhm_G = -fwhm_L*df_fwhm_G*g/f/5
  dx_fwhm_L = g - fwhm_L*df_fwhm_L*g/f/5
  deta_x = 0.33348*x**2-0.95438*x+1.36603
  return deta_x*dx_fwhm_G, deta_x*dx_fwhm_L


'''
mathfun:
subpackage for the mathematical functions for TRXASprefitpack

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from .peak_shape import voigt, edge_gaussian, edge_lorenzian
from .peak_shape import deriv_voigt, deriv_edge_gaussian, deriv_edge_lorenzian
from .peak_shape import voigt_thy, deriv_voigt_thy
from .irf import gau_irf, cauchy_irf, pvoigt_irf
from .irf import calc_eta, calc_fwhm, deriv_eta, deriv_fwhm
from .exp_conv_irf import exp_conv_gau, exp_conv_cauchy, exp_conv_pvoigt
from .exp_conv_irf import deriv_exp_conv_gau, deriv_exp_conv_cauchy
from .exp_conv_irf import deriv_exp_sum_conv_gau, deriv_exp_sum_conv_cauchy
from .exp_conv_irf import dmp_osc_conv_gau, dmp_osc_conv_cauchy, dmp_osc_conv_pvoigt
from .exp_conv_irf import deriv_dmp_osc_conv_gau, deriv_dmp_osc_conv_cauchy
from .rate_eq import solve_model, solve_seq_model, solve_l_model, compute_model
from .rate_eq import compute_signal_gau, compute_signal_cauchy
from .rate_eq import compute_signal_pvoigt
from .exp_decay_fit import exp_conv, fact_anal_exp_conv
from .exp_decay_fit import rate_eq_conv, fact_anal_rate_eq_conv
from .exp_decay_fit import dmp_osc_conv, fact_anal_dmp_osc_conv
from .exp_decay_fit import sum_exp_dmp_osc_conv, fact_anal_sum_exp_dmp_osc_conv


__all__ = ['voigt', 'edge_gaussian', 'edge_lorenzian', 'voigt_thy',
           'deriv_voigt', 'deriv_edge_gaussian', 'deriv_edge_lorenzian',
           'deriv_voigt_thy',
           'gau_irf', 'cauchy_irf', 'pvoigt_irf', 
           'calc_eta', 'calc_fwhm', 'deriv_eta', 'deriv_fwhm',
           'exp_conv_gau', 'exp_conv_cauchy', 'exp_conv_pvoigt',
           'deriv_exp_conv_gau', 'deriv_exp_conv_cauchy',
           'deriv_exp_sum_conv_gau', 'deriv_exp_sum_conv_cauchy',
           'dmp_osc_conv_gau', 'dmp_osc_conv_cauchy', 'dmp_osc_conv_pvoigt',
           'deriv_dmp_osc_conv_gau', 'deriv_dmp_osc_conv_cauchy',
           'solve_model', 'solve_seq_model', 'solve_l_model', 'compute_model',
           'compute_signal_gau', 'compute_signal_cauchy', 'compute_signal_pvoigt', 
           'exp_conv', 'fact_anal_exp_conv',
           'rate_eq_conv', 'fact_anal_rate_eq_conv',
           'dmp_osc_conv', 'fact_anal_dmp_osc_conv',
           'sum_exp_dmp_osc_conv', 'fact_anal_sum_exp_dmp_osc_conv']



'''
res_both:
submodule for residual function and gradient for fitting time delay scan with the
convolution of sum of (sum of exponential decay and damped oscillation) and instrumental response function

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Optional, Sequence, Tuple
import numpy as np
from ..mathfun.irf import calc_eta, deriv_eta
from ..mathfun.irf import calc_fwhm, deriv_fwhm
from ..mathfun.A_matrix import make_A_matrix_gau, make_A_matrix_cauchy
from ..mathfun.A_matrix import make_A_matrix_gau_osc, make_A_matrix_cauchy_osc, fact_anal_A
from ..mathfun.exp_conv_irf import deriv_exp_sum_conv_gau, deriv_exp_sum_conv_cauchy
from ..mathfun.exp_conv_irf import deriv_dmp_osc_sum_conv_gau, deriv_dmp_osc_sum_conv_cauchy

# residual and gradient function for exponential decay model + damped oscillation model

def residual_both(x0: np.ndarray, num_comp: int, num_comp_osc:int, base: bool, irf: str, 
                  t: Optional[Sequence[np.ndarray]] = None, 
                  intensity: Optional[Sequence[np.ndarray]] = None, 
                  eps: Optional[Sequence[np.ndarray]] = None) -> np.ndarray:
    '''
    residual_both
    `scipy.optimize.least_squares` compatible vector residual function for fitting multiple set of time delay scan with the
    sum of convolution of (sum of exponential decay damped oscillation) and instrumental response function  

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}` to :math:`2+N_{scan}+N_{\\tau}`: time constant of each decay component
        * :math:`2+N_{scan}+N_{\\tau}+i`: time constant of each damped oscillation
        * :math:`2+N_{scan}+N_{\\tau}+N_{osc}+i`: period of each damped oscillation

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}` to :math:`3+N_{scan}+N_{\\tau}`: time constant of each decay component
        * :math:`3+N_{scan}+N_{\\tau}+i`: time constant of each damped oscillation
        * :math:`3+N_{scan}+N_{\\tau}+N_{osc}+i`: period of each damped oscillation

     num_comp: number of exponential decay component (except base)
     num_comp_osc: number of damped oscillation component
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Residual vector
    
    Note:
     each dataset does not contain time range
    '''
    x0 = np.atleast_1d(x0)
    
    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
        eta = None
    else:
            num_irf = 2
            fwhm = calc_fwhm(x0[0], x0[1])
            eta = calc_eta(x0[0], x0[1])

    num_t0 = 0; sum = 0
    for d in intensity:
        num_t0 = d.shape[1] + num_t0
        sum = sum + d.size
    
    chi = np.empty(sum)

    tau = x0[num_irf+num_t0:num_irf+num_t0+num_comp]
    tau_osc = x0[num_irf+num_t0+num_comp:num_irf+num_t0+num_comp+num_comp_osc]
    period_osc = x0[num_irf+num_t0+num_comp+num_comp_osc:num_irf+num_t0+num_comp+2*num_comp_osc]

    if base:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau; k[-1] = 0
    else:
        k = 1/tau
    
    end = 0; t0_idx = num_irf
    for ti,d,e in zip(t,intensity,eps):
        A = np.empty((num_comp+1*base+2*num_comp_osc, d.shape[0]))
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A[:num_comp+1*base, :] = make_A_matrix_gau(ti-t0, fwhm, k)
                A[num_comp+1*base:, :] = make_A_matrix_gau_osc(ti-t0, fwhm, 1/tau_osc, period_osc)
            elif irf == 'c':
                A[:num_comp+1*base, :] = make_A_matrix_cauchy(ti-t0, fwhm, k)
                A[num_comp+1*base:, :] = make_A_matrix_cauchy_osc(ti-t0, fwhm, 1/tau_osc, period_osc)
            else:
                tmp_gau = make_A_matrix_gau(ti-t0, fwhm, k); 
                tmp_gau_osc = make_A_matrix_gau_osc(ti-t0, fwhm, 1/tau_osc, period_osc)
                tmp_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
                tmp_cauchy_osc = make_A_matrix_cauchy_osc(ti-t0, fwhm, 1/tau_osc, period_osc)
                A[:num_comp+1*base, :] = tmp_gau + eta*(tmp_cauchy-tmp_gau)
                A[num_comp+1*base:, :] = tmp_gau_osc + eta*(tmp_cauchy_osc-tmp_gau_osc)

            c = fact_anal_A(A, d[:,j], e[:,j])
            chi[end:end+d.shape[0]] = ((c@A) - d[:, j])/e[:, j]

            end = end + d.shape[0]
            t0_idx = t0_idx + 1
    return chi
    
def res_grad_both(x0: np.ndarray, num_comp: int, num_comp_osc:int, base: bool, irf: str, 
                   fix_param_idx: Optional[np.ndarray] = None,
                   t: Optional[Sequence[np.ndarray]] = None, 
                   intensity: Optional[Sequence[np.ndarray]] = None, 
                   eps: Optional[Sequence[np.ndarray]] = None) -> Tuple[np.ndarray, np.ndarray]:
    '''
    res_grad_both
    `scipy.optimize.minimize` compatible scalar residual and its gradient function for fitting multiple set of time delay scan with the
    sum of convolution of (sum of exponential decay damped oscillation) and instrumental response function  

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}` to :math:`2+N_{scan}+N_{\\tau}`: time constant of each decay component
        * :math:`2+N_{scan}+N_{\\tau}+i`: time constant of each damped oscillation
        * :math:`2+N_{scan}+N_{\\tau}+N_{osc}+i`: period of each damped oscillation

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}` to :math:`3+N_{scan}+N_{\\tau}`: time constant of each decay component
        * :math:`3+N_{scan}+N_{\\tau}+i`: time constant of each damped oscillation
        * :math:`3+N_{scan}+N_{\\tau}+N_{osc}+i`: period of each damped oscillation

     num_comp: number of exponential decay component (except base)
     num_comp_osc: number of damped oscillation component
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     fix_param_idx: idx for fixed parameter (masked array for `x0`)
     t: time points for each data set
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Tuple of scalar residual function :math:`(\\frac{1}{2}\\sum_i {res}^2_i)` and its gradient
    '''

    x0 = np.atleast_1d(x0)
    
    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
        eta = None
    else:
            num_irf = 2
            eta = calc_eta(x0[0], x0[1])
            fwhm = calc_fwhm(x0[0], x0[1])
            deta_G, deta_L = deriv_eta(x0[0], x0[1])
            dfwhm_G, dfwhm_L = deriv_fwhm(x0[0], x0[1])

    num_t0 = 0; sum = 0
    for d in intensity:
        num_t0 = d.shape[1] + num_t0
        sum = sum + d.size
    
    tau = x0[num_irf+num_t0:num_irf+num_t0+num_comp]
    tau_osc = x0[num_irf+num_t0+num_comp:num_irf+num_t0+num_comp+num_comp_osc]
    period_osc = x0[num_irf+num_t0+num_comp+num_comp_osc:num_irf+num_t0+num_comp+2*num_comp_osc]

    if base:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau; k[-1] = 0
    else:
        k = 1/tau
    
    num_param = num_irf+num_t0+num_comp+2*num_comp_osc

    chi = np.empty(sum)
    df = np.zeros((sum, num_irf+num_comp+2*num_comp_osc))
    grad = np.empty(num_param)
    
    end = 0; t0_idx = num_irf
    for ti,d,e in zip(t,intensity,eps):
        step = d.shape[0]
        A = np.empty((num_comp+1*base+2*num_comp_osc, step))
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A[:num_comp+1*base, :] = make_A_matrix_gau(ti-t0, fwhm, k)
                A[num_comp+1*base:, :] = make_A_matrix_gau_osc(ti-t0, fwhm, 1/tau_osc, period_osc)
            elif irf == 'c':
                A[:num_comp+1*base, :] = make_A_matrix_cauchy(ti-t0, fwhm, k)
                A[num_comp+1*base:, :] = make_A_matrix_cauchy_osc(ti-t0, fwhm, 1/tau_osc, period_osc)
            else:
                tmp_gau = make_A_matrix_gau(ti-t0, fwhm, k); 
                tmp_gau_osc = make_A_matrix_gau_osc(ti-t0, fwhm, 1/tau_osc, period_osc)
                tmp_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
                tmp_cauchy_osc = make_A_matrix_cauchy_osc(ti-t0, fwhm, 1/tau_osc, period_osc)
                diff = tmp_cauchy-tmp_gau; diff_osc = tmp_cauchy_osc-tmp_gau_osc
                A[:num_comp+1*base, :] = tmp_gau + eta*diff
                A[num_comp+1*base:, :] = tmp_gau_osc + eta*diff_osc

            c = fact_anal_A(A, d[:,j], e[:,j])
            chi[end:end+step] = (c@A-d[:,j])/e[:, j]

            c_amp = np.sqrt(c[num_comp+1*base:num_comp+1*base+num_comp_osc]**2+
            c[num_comp+1*base+num_comp_osc:]**2)
            phase_osc = -np.arctan2(c[num_comp+1*base+num_comp_osc:], 
            c[:num_comp+1*base+num_comp_osc])

            if irf == 'g':
                grad_decay = deriv_exp_sum_conv_gau(ti-t0, fwhm, 1/tau, c[:num_comp+1*base], base)
                grad_osc = deriv_dmp_osc_sum_conv_gau(ti-t0, fwhm, 1/tau_osc, period_osc, phase_osc, c_amp)
            elif irf == 'c':
                grad_decay = deriv_exp_sum_conv_cauchy(ti-t0, fwhm, 1/tau, c[:num_comp+1*base], base)
                grad_osc = deriv_dmp_osc_sum_conv_cauchy(ti-t0, fwhm, 1/tau_osc, period_osc, phase_osc, c_amp)
            else:
                grad_gau_decay = deriv_exp_sum_conv_gau(ti-t0, fwhm, 1/tau, c[:num_comp+1*base], base)
                grad_gau_osc = deriv_dmp_osc_sum_conv_gau(ti-t0, fwhm, 1/tau_osc, period_osc, phase_osc, c_amp)
                grad_cauchy_decay = deriv_exp_sum_conv_cauchy(ti-t0, fwhm, 1/tau, c[:num_comp+1*base], base)
                grad_cauchy_osc = deriv_dmp_osc_sum_conv_cauchy(ti-t0, fwhm, 1/tau_osc, period_osc, phase_osc, c_amp)
                grad_decay = grad_gau_decay + eta*(grad_cauchy_decay-grad_gau_decay)
                grad_osc = grad_gau_osc + eta*(grad_cauchy_osc-grad_gau_osc)
            
            grad_decay = np.einsum('i,ij->ij', 1/e[:, j], grad_decay)
            grad_osc = np.einsum('i,ij->ij', 1/e[:, j], grad_osc)
            grad_sum = grad_decay[:, :2] + grad_osc[:, :2]

            if irf in ['g', 'c']:
                df[end:end+step, 0] = grad_sum[:, 1]

            else:
                cdiff = (c[:num_comp+1*base]@diff + c[num_comp+1*base:]@diff_osc)/e[:, j]
                df[end:end+step, 0] = dfwhm_G*grad_sum[:, 1]+deta_G*cdiff
                df[end:end+step, 1] = dfwhm_L*grad_sum[:, 1]+deta_L*cdiff

            grad[t0_idx] = -chi[end:end+step]@grad_sum[:, 0]
            df[end:end+step, num_irf:num_irf+num_comp] = \
                np.einsum('j,ij->ij', -1/tau**2, grad_decay[:, 2:])
            df[end:end+step, num_irf+num_comp:num_irf+num_comp+num_comp_osc] = \
                np.einsum('j,ij->ij',-1/tau_osc**2, grad_osc[:, 2:2+num_comp_osc])
            df[end:end+step, num_irf+num_comp+num_comp_osc:] = grad_osc[:, 2+num_comp_osc:2+2*num_comp_osc]


            end = end + step
            t0_idx = t0_idx + 1
    
    mask = np.ones(num_param, dtype=bool)
    mask[num_irf:num_irf+num_t0] = False
    grad[mask] = chi@df

    if fix_param_idx is not None:
        grad[fix_param_idx] = 0
    

    return np.sum(chi**2)/2, grad


'''
param_bound:
submodule for setting default parameter boundary of
irf parameter, time zero and lifetime constant tau
:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''
from typing import Tuple, Union
import numpy as np


def set_bound_t0(t0: float, fwhm: Union[float, np.ndarray]) -> Tuple[float, float]:
    '''
    Setting bound for time zero

    Args:
     t0: initial guess for time zero
     fwhm: initial guess for full width at half maximum of instrumental response function
           `float` for gaussian and cauchy shape, `np.ndarray` with two element `(fwhm_G, fwhm_L)` for pseudo voigt shape.
    
    Returns:
     Tuple of upper and lower bound of time zero
    '''
    if not isinstance(fwhm, np.ndarray):
        bound = (t0-2*fwhm, t0+2*fwhm)
    else:
        fwhm_eff = 0.5346*fwhm[1] + \
                np.sqrt(0.2166*fwhm[1]**2+fwhm[0]**2)
        bound = (t0-2*fwhm_eff, t0+2*fwhm_eff)
    
    return bound

def set_bound_e0(e0: float, fwhm_G: float, fwhm_L: float) -> Tuple[float, float]:
    '''
    Setting bound for peak position and edge position

    Args:
     e0: initial guess for peak position
     fwhm_G: initial guess of fwhm_G parameter of voigt component
     fwhm_L: initial guess of fwhm_L parameter of voigt component
    
    Returns:
     Tuple of upper and lower bound of peak position
    '''
    if fwhm_G == 0:
        bound = (e0-fwhm_L/2, e0+fwhm_L/2)
    elif fwhm_L == 0:
        bound = (e0-fwhm_G/2, e0+fwhm_G/2)
    else:
        fwhm_eff = 0.5346*fwhm_L+np.sqrt(0.2166*fwhm_L**2+fwhm_G**2)
        bound = (e0-fwhm_eff/2, e0+fwhm_eff/2)
    
    return bound


def set_bound_tau(tau: float, fwhm: Union[float, np.ndarray]) -> Tuple[float, float]:
    '''
    Setting bound for lifetime constant

    Args:
      tau: initial guess for lifetime constant
      fwhm: initial guess for full width at half maximum of instrumental response function
           `float` for gaussian and cauchy shape, `np.ndarray` with two element `(fwhm_G, fwhm_L)` for pseudo voigt shape.

    Returns:
     Tuple of upper bound and lower bound of tau
    '''
    if not isinstance(fwhm, np.ndarray):
        fwhm_eff = fwhm
    else:
        fwhm_eff = 0.5346*fwhm[1] + \
                np.sqrt(0.2166*fwhm[1]**2+fwhm[0]**2)


    bound = (tau/2, 2*fwhm_eff)
    if fwhm_eff <= tau < 4*fwhm_eff:
        bound = (fwhm_eff/2, 8*fwhm_eff)
    elif 4*fwhm_eff <= tau < 16*fwhm_eff:
        bound = (2*fwhm_eff, 32*fwhm_eff)
    elif 16*fwhm_eff <= tau < 64*fwhm_eff:
        bound = (8*fwhm_eff, 128*fwhm_eff)
    elif 64*fwhm_eff <= tau < 256*fwhm_eff:
        bound = (32*fwhm_eff, 512*fwhm_eff)
    elif 256*fwhm_eff <= tau < 1024*fwhm_eff:
        bound = (128*fwhm_eff, 2048*fwhm_eff)
    elif 1024*fwhm_eff <= tau < 4096*fwhm_eff:
        bound = (512*fwhm_eff, 8192*fwhm_eff)
    elif 4096*fwhm_eff <= tau < 16384*fwhm_eff:
        bound = (2048*fwhm_eff, 32768*fwhm_eff)
    elif 16384*fwhm_eff <= tau < 65536*fwhm_eff:
        bound = (8192*fwhm_eff, 131072*fwhm_eff)
    elif 65536*fwhm_eff <= tau < 262144*fwhm_eff:
        bound = (32768*fwhm_eff, 524288*fwhm_eff)
    elif tau >= 262144*fwhm_eff:
        bound = (131072*fwhm_eff, 2*tau)
    return bound


import os
import sys
import unittest
import numpy as np

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack import calc_eta, calc_fwhm, deriv_eta, deriv_fwhm
from TRXASprefitpack.mathfun.deriv_tst import test_num_deriv
from TRXASprefitpack import exp_conv_gau, exp_conv_cauchy
from TRXASprefitpack import deriv_exp_conv_gau, deriv_exp_conv_cauchy
from TRXASprefitpack import deriv_exp_sum_conv_gau, deriv_exp_sum_conv_cauchy

class TestDerivExpConvIRF(unittest.TestCase):
    def test_deriv_exp_conv_gau_1(self):
        '''
        Test gradient of convolution of exponenetial decay and gaussian irf (tau: 1, fwhm: 0.15)
        '''
        tau = 1
        fwhm = 0.15
        t = np.linspace(-1, 100, 2001)
        tst = deriv_exp_conv_gau(t, fwhm, 1/tau)
        ref = test_num_deriv(exp_conv_gau, t, fwhm, 1/tau)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)
    
    def test_deriv_exp_sum_conv_gau_1(self):
        '''
        Test gradient of convolution of sum of exponenetial decay and gaussian irf (tau1: 1, tau2: 100, fwhm: 0.15, base: True)
        '''
        tau_1 = 1
        tau_2 = 100
        k = np.array([1/tau_1, 1/tau_2])
        fwhm = 0.15
        base = True
        t = np.linspace(-1, 1000, 20001)
        c = np.array([1, 1, 1])
        tst = deriv_exp_sum_conv_gau(t, fwhm, k, c, base)
        ref = test_num_deriv(lambda t, fwhm, k1, k2: exp_conv_gau(t, fwhm, k1)+exp_conv_gau(t, fwhm, k2)+exp_conv_gau(t, fwhm, 0),
        t, fwhm, 1/tau_1, 1/tau_2)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)

    def test_deriv_exp_sum_conv_gau_2(self):
        '''
        Test gradient of convolution of sum of exponenetial decay and gaussian irf (tau1: 1, tau2: 100, fwhm: 0.15, base: False)
        '''
        tau_1 = 1
        tau_2 = 100
        k = np.array([1/tau_1, 1/tau_2])
        fwhm = 0.15
        base = False
        t = np.linspace(-1, 1000, 20001)
        c = np.array([1, 1, 1])
        tst = deriv_exp_sum_conv_gau(t, fwhm, k, c, base)
        ref = test_num_deriv(lambda t, fwhm, k1, k2: exp_conv_gau(t, fwhm, k1)+exp_conv_gau(t, fwhm, k2),
        t, fwhm, 1/tau_1, 1/tau_2)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)

    def test_deriv_exp_conv_cauchy(self):
        '''
        Test gradient of convolution of exponenetial decay and cauchy irf (tau: 1, fwhm: 0.15)
        '''
        tau = 1
        fwhm = 0.15
        t = np.linspace(-1, 100, 2001)
        tst = deriv_exp_conv_cauchy(t, fwhm, 1/tau)
        ref = test_num_deriv(exp_conv_cauchy, t, fwhm, 1/tau)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)
    
    def test_deriv_exp_sum_conv_cauchy_1(self):
        '''
        Test gradient of convolution of sum of exponenetial decay and cauchy irf (tau1: 1, tau2: 100, fwhm: 0.15, base: True)
        '''
        tau_1 = 1
        tau_2 = 100
        k = np.array([1/tau_1, 1/tau_2])
        fwhm = 0.15
        base = True
        t = np.linspace(-1, 1000, 20001)
        c = np.array([1, 1, 1])
        tst = deriv_exp_sum_conv_cauchy(t, fwhm, k, c, base)
        ref = test_num_deriv(lambda t, fwhm, k1, k2: exp_conv_cauchy(t, fwhm, k1)+
        exp_conv_cauchy(t, fwhm, k2)+exp_conv_cauchy(t, fwhm, 0),
        t, fwhm, 1/tau_1, 1/tau_2)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)

    def test_deriv_exp_sum_conv_cauchy_2(self):
        '''
        Test gradient of convolution of sum of exponenetial decay and gaussian irf (tau1: 1, tau2: 100, fwhm: 0.15, base: False)
        '''
        tau_1 = 1
        tau_2 = 100
        k = np.array([1/tau_1, 1/tau_2])
        fwhm = 0.15
        base = False
        t = np.linspace(-1, 1000, 20001)
        c = np.array([1, 1])
        tst = deriv_exp_sum_conv_cauchy(t, fwhm, k, c, base)
        ref = test_num_deriv(lambda t, fwhm, k1, k2: exp_conv_cauchy(t, fwhm, k1)+
        exp_conv_cauchy(t, fwhm, k2),
        t, fwhm, 1/tau_1, 1/tau_2)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)
    
    def test_deriv_exp_conv_pvoigt(self):
        '''
        Test gradient of convolution of exponenetial decay and pseudo voigt irf (tau: 1, fwhm_G: 0.1, fwhm_L: 0.15)
        Note.
         Not implemented in mathfun module, check implemenetation in res_grad_decay function
        '''

        def tmp_fun(t, fwhm_G, fwhm_L, k):
            fwhm = calc_fwhm(fwhm_G, fwhm_L)
            eta = calc_eta(fwhm_G, fwhm_L)
            gau = exp_conv_gau(t, fwhm, k)
            cauchy = exp_conv_cauchy(t, fwhm, k)
            return gau + eta*(cauchy-gau)
        
        tau_1 = 1
        fwhm_G = 0.1
        fwhm_L = 0.15
        t = np.linspace(-1, 200, 2001)

        grad = np.empty((t.size, 4))
        fwhm = calc_fwhm(fwhm_G, fwhm_L)
        eta = calc_eta(fwhm_G, fwhm_L)
        dfwhm_G, dfwhm_L = deriv_fwhm(fwhm_G, fwhm_L)
        deta_G, deta_L = deriv_eta(fwhm_G, fwhm_L)
        diff = exp_conv_cauchy(t, fwhm, 1/tau_1) - \
            exp_conv_gau(t, fwhm, 1/tau_1)
        grad_gau = deriv_exp_conv_gau(t, fwhm, 1/tau_1)
        grad_cauchy = deriv_exp_conv_cauchy(t, fwhm, 1/tau_1)
        grad_tot = grad_gau + eta*(grad_cauchy-grad_gau)
        grad[:, 0] = grad_tot[:, 0]; grad[:, 3] = grad_tot[:, 2]
        grad[:, 1] = dfwhm_G*grad_tot[:, 1] + deta_G*diff
        grad[:, 2] = dfwhm_L*grad_tot[:, 1] + deta_L*diff

        ref = test_num_deriv(tmp_fun, t, fwhm_G, fwhm_L, 1/tau_1)

        result = np.allclose(grad, ref)
        self.assertEqual(result, True)



if __name__ == '__main__':
    unittest.main()



# match scale
# match scaling of each energy scan data via one reference time scan data

import argparse
import numpy as np

from ..mathfun.irf import calc_fwhm, calc_eta
from ..mathfun import fact_anal_exp_conv
from ..mathfun.A_matrix import make_A_matrix_exp
from .misc import read_data

description = '''
match scale: match scaling of each energy scan data to one reference time delay scan data.
Experimentally measured time delay scan data has unsystematic error, which makes correct scaling of
each energy scan data ambiguous. To reduce such ambiguity, it fits reference time delay scan with the sum of
the convolution of exponential decay and instrumental response function.  
'''

epilog = '''
*Note

1. Fitting parameters (time_zero, fwhm, tau) are should be evaluated previosuly from fit_tscan utility.
2. Time zero of reference time delay scan and energy scan should be same.
3. Energy scan range contains energy of reference time delay scan.
4. if you set shape of irf to pseudo voigt (pv), then you should provide two full width at half maximum
   value for gaussian and cauchy parts, respectively.
'''

irf_help = '''
shape of instrument response functon
g: gaussian distribution
c: cauchy distribution
pv: pseudo voigt profile, linear combination of gaussian distribution and cauchy distribution 
    pv = eta*c+(1-eta)*g 
    the uniform fwhm parameter and 
    mixing parameter are determined according to Journal of Applied Crystallography. 33 (6): 1311–1316. 
'''

fwhm_G_help = '''
full width at half maximum for gaussian shape
It would not be used when you set cauchy irf function
'''

fwhm_L_help = '''
full width at half maximum for cauchy shape
It would not be used when you did not set irf or use gaussian irf function
'''


def match_scale():

    tmp = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=tmp,
                                     description=description,
                                     epilog=epilog)
    parser.add_argument('--irf', default='g', choices=['g', 'c', 'pv'],
                        help=irf_help)
    parser.add_argument('--fwhm_G', type=float,
                        help=fwhm_G_help)
    parser.add_argument('--fwhm_L', type=float,
                        help=fwhm_L_help)
    parser.add_argument('prefix',
                        help='prefix for energy scan files ' +
                        'It will read prefix_i.txt')
    parser.add_argument('--ref_tscan_energy', type=float,
    help='energy of reference time delay scan')
    parser.add_argument('--ref_tscan_file',
    help='filename for reference time delay scan data')
    parser.add_argument('--escan_time', type=float, nargs='+',
                         help='time points for energy scan data')
    parser.add_argument('-t0', '--time_zero', type=float,
                        help='time zero for reference time scan')
    parser.add_argument('--tau', type=float, nargs='*',
                        help='lifetime of each component')
    parser.add_argument('--no_base', action='store_false',
                        help='exclude baseline')
    parser.add_argument('-o', '--out', default='out',
                        help='prefix for scaled energy scan and error')
    args = parser.parse_args()

    prefix = args.prefix
    out_prefix = args.out

    irf = args.irf
    if irf == 'g':
        if args.fwhm_G is None:
            raise Exception('You are using gaussian irf, so you should set fwhm_G!\n')
        else:
            fwhm = args.fwhm_G
            eta = 0
    elif irf == 'c':
        if args.fwhm_L is None:
            raise Exception('You are using cauchy/lorenzian irf,' + 
            'so you should set fwhm_L!\n')
        else:
            fwhm = args.fwhm_L
    else:
        if (args.fwhm_G is None) or (args.fwhm_L is None):
            raise Exception('You are using pseudo voigt irf,' +
            'so you should set both fwhm_G and fwhm_L!\n')
        else:
            fwhm = calc_fwhm(args.fwhm_G, args.fwhm_L)
            eta = calc_eta(args.fwhm_G, args.fwhm_L)

    if args.tau is None:
        base = True
    else:
        tau = np.array(args.tau)
        base = args.no_base

    if (args.time_zero is None):
        print('You should set time_zeros!\n')
        return
    else:
        time_zero = args.time_zero
    
    ref_tscan_energy = args.ref_tscan_energy
    ref_tscan_data = np.genfromtxt(args.ref_tscan_file)
    escan_time = np.array(args.escan_time)
    e = np.genfromtxt(f'{prefix}_1.txt')[:, 0]
    escan_data, escan_eps = read_data(prefix, escan_time.size, e.size, 10)
    print(f'It reads {escan_time.size} number of energy scan!\n')

    escan_data_scaled = np.empty((escan_data.shape[0], escan_data.shape[1]+1))
    escan_data_scaled[:, 0] = e
    e_ref_idx = np.argwhere(e == ref_tscan_energy)[0][0]

    c = fact_anal_exp_conv(ref_tscan_data[:,0]-time_zero, fwhm, tau, base, irf, eta,
    intensity=ref_tscan_data[:, 1], eps=ref_tscan_data[:, 2])
    A_slec = make_A_matrix_exp(escan_time-time_zero, fwhm, tau, base, irf)
    fit_slec = c@A_slec
    sample_e = escan_data[e_ref_idx, :]
    scale_factor = fit_slec/sample_e
    escan_data_scaled[:, 1:] = np.einsum('j,ij->ij', scale_factor, escan_data)
    escan_eps_scaled = np.einsum('j,ij->ij', scale_factor, escan_eps)

    header_escan = '\t'.join(list(map(str,escan_time)))
    

    np.savetxt(f'{out_prefix}_escan_scaled.txt', escan_data_scaled, header='energy \t'+header_escan)
    np.savetxt(f'{out_prefix}_eps_scaled.txt', escan_eps_scaled, fmt=len(escan_time)*['%.8e'], header=header_escan)
    np.savetxt(f'{out_prefix}_escan_time.txt', escan_time)

    return


'''
res_osc:
submodule for residual function and gradient for fitting time delay scan with the
convolution of sum of damped oscillation and instrumental response function 

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Tuple, Optional, Sequence
import numpy as np
from ..mathfun.irf import calc_eta, deriv_eta
from ..mathfun.irf import calc_fwhm, deriv_fwhm
from ..mathfun.A_matrix import make_A_matrix_gau_osc, make_A_matrix_cauchy_osc, fact_anal_A
from ..mathfun.exp_conv_irf import deriv_dmp_osc_sum_conv_gau, deriv_dmp_osc_sum_conv_cauchy

# residual and gradient function for damped oscillation model 

def residual_dmp_osc(x0: np.ndarray, num_comp: int, irf: str, 
                    t: Optional[Sequence[np.ndarray]] = None, 
                    intensity: Optional[Sequence[np.ndarray]] = None, eps: Optional[Sequence[np.ndarray]] = None) -> np.ndarray:
    '''
    residual_dmp_osc
    `scipy.optimize.least_squares` compatible gradient of vector residual function for fitting multiple set of time delay scan with the
    sum of convolution of damped oscillation and instrumental response function  

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}+i`: time constant of each damped oscillation
        * :math:`2+N_{scan}+N_{osc}+i`: period of each damped oscillation

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}+i`: time constant of each damped oscillation
        * :math:`3+N_{scan}+N_{osc}+i`: period of each damped oscillation
           
     num_comp: number of damped oscillation component
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets
    
    Returns:
     Residual vector
    '''

    x0 = np.atleast_1d(x0)
    
    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
    else:
            num_irf = 2
            fwhm = calc_fwhm(x0[0], x0[1])
            eta = calc_eta(x0[0], x0[1])

    num_t0 = 0; sum = 0
    for d in intensity:
        num_t0 = d.shape[1] + num_t0
        sum = sum + d.size
    
    chi = np.empty(sum)
    
    tau = x0[num_irf+num_t0:num_irf+num_t0+num_comp]
    period = x0[num_irf+num_t0+num_comp:num_irf+num_t0+2*num_comp]

    end = 0; t0_idx = num_irf
    for ti,d,e in zip(t,intensity,eps):
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A = make_A_matrix_gau_osc(ti-t0, fwhm, 1/tau, period)
            elif irf == 'c':
                A = make_A_matrix_cauchy_osc(ti-t0, fwhm, 1/tau, period)
            else:
                A_gau = make_A_matrix_gau_osc(ti-t0, fwhm, 1/tau, period)
                A_cauchy = make_A_matrix_cauchy_osc(ti-t0, fwhm, 1/tau, period)
                A = A_gau + eta*(A_cauchy-A_gau)
            c = fact_anal_A(A, d[:,j], e[:,j])
            chi[end:end+d.shape[0]] = ((c@A) - d[:, j])/e[:, j]

            end = end + d.shape[0]
            t0_idx = t0_idx + 1

    return chi
    
def res_grad_dmp_osc(x0: np.ndarray, num_comp: int, irf: str, 
                     fix_param_idx: Optional[np.ndarray] = None,
                     t: Optional[Sequence[np.ndarray]] = None, 
                     intensity: Optional[Sequence[np.ndarray]] = None, 
                     eps: Optional[Sequence[np.ndarray]] = None) -> Tuple[np.ndarray, np.ndarray]:
    '''
    res_grad_dmp_osc
    scipy.optimize.minimize compatible pair of scalar residual function and its gradient for fitting multiple set of time delay scan with the
    sum of convolution of damped oscillation and instrumental response function  

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}+i`: time constant of each damped oscillation
        * :math:`2+N_{scan}+N_{osc}+i`: period of each damped oscillation

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}+i`: time constant of each damped oscillation
        * :math:`3+N_{scan}+N_{osc}+i`: period of each damped oscillation
           
     num_comp: number of damped oscillation component
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     fix_param_idx: index for fixed parameter (masked array for `x0`)
     t: time points for each data set
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Tuple of scalar residual function :math:`(\\frac{1}{2}\\sum_i {res}^2_i)` and its gradient
    '''
    x0 = np.atleast_1d(x0)
    
    if irf in ['g', 'c']:
            num_irf = 1 
            fwhm = x0[0]
    else:
        num_irf = 2
        eta = calc_eta(x0[0], x0[1])
        fwhm = calc_eta(x0[0], x0[1])
        dfwhm_G, dfwhm_L = deriv_fwhm(x0[0], x0[1])
        deta_G, deta_L = deriv_eta(x0[0], x0[1])

    num_t0 = 0; sum = 0
    for d in intensity:
        num_t0 = num_t0 + d.shape[1]
        sum = sum + d.size

    tau = x0[num_irf+num_t0:num_irf+num_t0+num_comp]
    period = x0[num_irf+num_t0+num_comp:num_irf+num_t0+2*num_comp]
    
    num_param = num_irf+num_t0+2*num_comp
    chi = np.empty(sum)
    df = np.zeros((sum, num_irf+2*num_comp))
    grad = np.empty(num_param)

    end = 0; t0_idx = num_irf

    for ti,d,e in zip(t, intensity, eps):
        step = d.shape[0]
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A = make_A_matrix_gau_osc(ti-t0, fwhm, 1/tau, period)
            elif irf == 'c':
                A = make_A_matrix_cauchy_osc(ti-t0, fwhm, 1/tau, period)
            else:
                A_gau = make_A_matrix_gau_osc(ti-t0, fwhm, 1/tau, period)
                A_cauchy = make_A_matrix_cauchy_osc(ti-t0, fwhm, 1/tau, period)
                diff = A_cauchy-A_gau
                A = A_gau + eta*diff
            c = fact_anal_A(A, d[:,j], e[:,j])
            chi[end:end+step] = (c@A-d[:,j])/e[:, j]

            c_amp = np.sqrt(c[:num_comp]**2+c[num_comp:]**2)
            phase = -np.arctan2(c[num_comp:], c[:num_comp])
                
            if irf == 'g':
                grad_tmp = deriv_dmp_osc_sum_conv_gau(ti-t0, fwhm, 1/tau, period, phase, c_amp)
            elif irf == 'c':
                grad_tmp = deriv_dmp_osc_sum_conv_cauchy(ti-t0, fwhm, 1/tau, period, phase, c_amp)
            else:
                grad_gau = deriv_dmp_osc_sum_conv_gau(ti-t0, fwhm[0], 1/tau, period, phase, c_amp)
                grad_cauchy = deriv_dmp_osc_sum_conv_cauchy(ti-t0, fwhm[1], 1/tau, period, phase, c_amp)
                grad_tmp = grad_gau + eta*(grad_cauchy-grad_gau)
   
            grad_tmp = np.einsum('i,ij->ij', 1/e[:, j], grad_tmp)
            if irf in ['g', 'c']:
                df[end:end+step, 0] = grad_tmp[:, 1]
            else:
                cdiff = (c@diff)/e[:, j]
                df[end:end+step, 0] = dfwhm_G*grad_tmp[:, 1]+deta_G*cdiff
                df[end:end+step, 1] = dfwhm_L*grad_tmp[:, 1]+deta_L*cdiff

            grad[t0_idx] = -chi[end:end+step]@grad_tmp[:, 0]
            df[end:end+step, :num_irf+num_comp] = \
                np.einsum('j,ij->ij', -1/tau**2, grad_tmp[:, 2:2+num_comp])
            df[end:end+step, num_irf+num_comp:] = grad_tmp[:, 2+num_comp:2+2*num_comp]

                
            end = end + step
            t0_idx = t0_idx + 1
    
    mask = np.ones(num_param, dtype=bool)
    mask[num_irf:num_irf+num_t0] = False
    grad[mask] = chi@df

    if fix_param_idx is not None:
        df[:, fix_param_idx] = 0

    return np.sum(chi**2)/2, grad


'''
exp_conv_irf:
submodule for the mathematical functions for
exponential decay convolved with irf

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Union, Optional, Tuple
import numpy as np
from scipy.special import erfc, erfcx, wofz, exp1

# special function: scaled expoential integral
# if scipy starts to support scaled expoential integral I will delete this function 

def exp1x_asymp(z: Union[complex, np.ndarray]) -> Union[complex, np.ndarray]:
    return z*(1+z*(1+2*z*(1+3*z*(1+4*z*(1+5*z*(1+6*z*(1+7*z*(1+8*z*(1+9*z*(1+10*z))))))))))

def exp1x(z: Union[complex, np.ndarray]) -> Union[complex, np.ndarray]:
    
    if not isinstance(z, np.ndarray):
            if np.abs(z.real) < 200:
                # ans = np.exp(z)*exp1(z)
                ans = complex(np.cos(z.imag), np.sin(z.imag))*exp1(z)
                ans = np.exp(z.real)*ans
            else:
                ans = -exp1x_asymp(-1/z)
    else:
        mask = np.abs(z.real) < 200
        inv_mask = np.invert(mask)
        ans = np.empty(z.size, dtype=np.complex128)
        
        # abs(Re z) < 200
        ans[mask] = (np.cos(z[mask].imag)+complex(0,1)*np.sin(z[mask].imag))*exp1(z[mask])
        ans[mask] = np.exp(z.real[mask])*ans[mask]

        # abs(Re z) > 200, use asymptotic series
        ans[inv_mask] = -exp1x_asymp(-1/z[inv_mask])
    return ans

# calculate convolution of exponential decay and instrumental response function

def exp_conv_gau(t: Union[float, np.ndarray], fwhm: float,
                 k: float) -> Union[float, np.ndarray]:

    '''
    Compute exponential function convolved with normalized gaussian
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution
      k: rate constant (inverse of life time)

    Returns:
     Convolution of normalized gaussian distribution and exponential
     decay :math:`(\\exp(-kt))`.
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    ksigma = k*sigma
    z = ksigma - t/sigma

    if not isinstance(t, np.ndarray):
        if z > 0:
            ans = 1/2*np.exp(-(t/sigma)**2/2) * \
                erfcx(z/np.sqrt(2))
        else:
            ans = 1/2*np.exp(ksigma*z-ksigma**2/2) * \
                erfc(z/np.sqrt(2))
    else:
        mask = z > 0
        inv_mask = np.invert(mask)
        ans = np.empty(t.size, dtype=np.float64)
        ans[mask] = 1/2*np.exp(-(t[mask]/sigma)**2/2) * \
            erfcx(z[mask]/np.sqrt(2))
        ans[inv_mask] = 1/2*np.exp(ksigma*z[inv_mask]-ksigma**2/2) * \
            erfc(z[inv_mask]/np.sqrt(2))

    return ans

def exp_conv_cauchy(t: Union[float, np.ndarray],
                    fwhm: float,
                    k: float) -> Union[float, np.ndarray]:

    '''
    Compute exponential function convolved with normalized cauchy
    distribution

    Args:
       t: time
       fwhm: full width at half maximum of cauchy distribution
       k: rate constant (inverse of life time)

    Returns:
      Convolution of normalized cauchy distribution and
      exponential decay :math:`(\\exp(-kt))`.
    '''

    if k == 0:
        ans = 0.5+1/np.pi*np.arctan(2*t/fwhm)
    else:
        z = -k*t - complex(0, k*fwhm/2)
        ans = exp1x(z).imag/np.pi
    return ans

def exp_conv_pvoigt(t: Union[float, np.ndarray],
                    fwhm: float,
                    eta: float,
                    k: float) -> Union[float, np.ndarray]:

    '''
    Compute exponential function convolved with normalized pseudo
    voigt profile (i.e. linear combination of normalized gaussian and
    cauchy distribution)

    :math:`\\eta C(\\mathrm{fwhm}, t) + (1-\\eta)G(\\mathrm{fwhm}, t)`

    Args:
       t: time
       fwhm: full width at half maximum parameter for pseudo voigt irf
       eta: mixing parameter
       k: rate constant (inverse of life time)

    Returns:
       Convolution of normalized pseudo voigt profile and
       exponential decay :math:`(\\exp(-kt))`.
    '''

    u = exp_conv_gau(t, fwhm, k)
    v = exp_conv_cauchy(t, fwhm, k)

    return u + eta*(v-u)

# calculate derivative of the convolution of exponential decay and instrumental response function

def deriv_exp_conv_gau(t: Union[float, np.ndarray], fwhm: float,
                 k: float) -> np.ndarray:

    '''
    Compute derivative of exponential function convolved with normalized gaussian
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution
      k: rate constant (inverse of life time)

    Returns:
     Derivative of Convolution of normalized gaussian distribution and exponential
     decay :math:`(\\exp(-kt))`.
    Note:

     * 1st column: df/dt
     * 2nd column: df/d(fwhm)
     * 3rd column: df/dk

    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    f = exp_conv_gau(t, fwhm, k)
    g = np.exp(-(t/sigma)**2/2)/np.sqrt(2*np.pi)

    if not isinstance(t, np.ndarray):
        grad = np.empty(3)
        grad[0] = g/sigma - k*f
        grad[1] = ((k**2*sigma)*f - (k+t/sigma**2)*g)/(2*np.sqrt(2*np.log(2)))
        grad[2] = (sigma**2*k-t)*f - sigma*g
    else:
        grad = np.empty((t.size, 3))
        grad[:, 0] = g/sigma - k*f
        grad[:, 1] = ((k**2*sigma)*f - (k+t/sigma**2)*g)/(2*np.sqrt(2*np.log(2)))
        grad[:, 2] = (sigma**2*k-t)*f - sigma*g

    return grad

def deriv_exp_conv_cauchy(t: Union[float, np.ndarray],
                    fwhm: float,
                    k: float) -> Union[float, np.ndarray]:

    '''
    Compute derivative of the convolution of exponential function and normalized cauchy
    distribution

    Args:
       t: time
       fwhm: full width at half maximum of cauchy distribution
       k: rate constant (inverse of life time)

    Returns:
      Derivative of convolution of normalized cauchy distribution and
      exponential decay :math:`(\\exp(-kt))`.
    
    Note:

     * 1st column: df/dt
     * 2nd column: df/d(fwhm)
     * 3rd column: df/dk
    '''

    if not isinstance(t, np.ndarray):
        grad = np.empty(3)
        if k == 0:
            grad[0] = 2/(np.pi*fwhm*(1+(2*t/fwhm)**2))
            grad[1] = -t/fwhm*grad[0]
            grad[2] = 0
        else:
            z = (t+complex(0, fwhm/2))
            f = exp1x(-k*z)
            tmp = -(k*f+1/z)/np.pi
            grad[0] = tmp.imag
            grad[1] = tmp.real/2
            grad[2] = -(t*f.imag+fwhm*f.real/2)/np.pi
    else:
        grad = np.empty((t.size, 3))
        if k == 0:
            grad[:, 0] = 2/(np.pi*fwhm*(1+(2*t/fwhm)**2))
            grad[:, 1] = -t/fwhm*grad[:, 0]
            grad[:, 2] = 0
        else:
            z = (t+complex(0, fwhm/2))
            f = exp1x(-k*z)
            tmp = -(k*f+1/z)/np.pi
            grad[:, 0] = tmp.imag
            grad[:, 1] = tmp.real/2
            grad[:, 2] = -(t*f.imag+fwhm*f.real/2)/np.pi
    return grad

# calculate derivative of the convolution of sum of exponential decay and instrumental response function

def deriv_exp_sum_conv_gau(t: np.ndarray, fwhm: float,
                 k: np.ndarray, c: np.ndarray, base: Optional[bool] = True) -> np.ndarray:

    '''
    Compute derivative of sum of exponential function convolved with normalized gaussian
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution
      k: rate constant (inverse of life time)
      c: coefficient
      base: include baseline (i.e. k=0)

    Returns:
     Derivative of Convolution of normalized gaussian distribution and sum of exponential
     decay :math:`(\\exp(-kt))`.
    Note:

     * 1st column: df/dt
     * 2nd column: df/d(fwhm)
     * 2+i th column: df/dk_i
    '''
    grad = np.zeros((t.size, 2+k.size))
    for i in range(k.size):
        grad_i = deriv_exp_conv_gau(t, fwhm, k[i])
        grad[:, 0] = grad[:, 0] + c[i]*grad_i[:, 0]
        grad[:, 1] = grad[:, 1] + c[i]*grad_i[:, 1]
        grad[:, 2+i] = c[i]*grad_i[:, 2]
    if base:
        grad_i = deriv_exp_conv_gau(t, fwhm, 0)
        grad[:, 0] = grad[:, 0] + c[-1]*grad_i[:, 0]
        grad[:, 1] = grad[:, 1] + c[-1]*grad_i[:, 1]

    return grad

def deriv_exp_sum_conv_cauchy(t: np.ndarray, fwhm: float,
                 k: np.ndarray, c: np.ndarray, base: Optional[bool] = True) -> np.ndarray:

    '''
    Compute derivative of sum of exponential function convolved with normalized cauchy
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of cauchy distribution
      k: rate constant (inverse of life time)
      c: coefficient
      base: include baseline (i.e. k=0)

    Returns:
     Derivative of Convolution of normalized cauchy distribution and sum of exponential
     decay :math:`(\\exp(-kt))`.
    Note:

     * 1st column: df/dt
     * 2nd column: df/d(fwhm)
     * 2+i th column: df/dk_i
    '''
    grad = np.zeros((t.size, 2+k.size))
    for i in range(k.size):
        grad_i = deriv_exp_conv_cauchy(t, fwhm, k[i])
        grad[:, 0] = grad[:, 0] + c[i]*grad_i[:, 0]
        grad[:, 1] = grad[:, 1] + c[i]*grad_i[:, 1]
        grad[:, 2+i] = c[i]*grad_i[:, 2]
    if base:
        grad_i = deriv_exp_conv_cauchy(t, fwhm, 0)
        grad[:, 0] = grad[:, 0] + c[-1]*grad_i[:, 0]
        grad[:, 1] = grad[:, 1] + c[-1]*grad_i[:, 1]

    return grad

def exp_mod_gau_cplx(t: Union[float, np.ndarray], sigma: float,
                     kr: float, ki: float) -> \
                        Union[Tuple[float, float], Tuple[np.ndarray, np.ndarray]]:
    '''
    Complex version of exponentially (exp(-krt+i*kit)) modified gaussian function

    Args:
     t: time
     sigma: deviation of gaussian distribution
     kr: real part of decay constant
     ki: imag part of decay constant
    
    Returns:
     complex version of exponentially modified gaussian distribution with mean = 0
    
    Note:
     first element is cosine and second element is sine.
    '''

    isigmak_cplx = complex(ki*sigma, kr*sigma)
    iz = isigmak_cplx - complex(0,1)*t/sigma

    if not isinstance(t, np.ndarray):
        if iz.imag > 0:
            ans = 1/2*np.exp(-(t/sigma)**2/2) * \
                wofz(iz/np.sqrt(2))
        else:
            ans = np.exp(isigmak_cplx**2/2-isigmak_cplx*iz) - \
                1/2*np.exp(-(t/sigma)**2/2)*wofz(-iz/np.sqrt(2)) 
    else:
        mask = iz.imag > 0
        inv_mask = np.invert(mask)
        ans = np.empty(t.size, dtype=np.complex128)
        ans[mask] = 1/2*np.exp(-(t[mask]/sigma)**2/2) * \
            wofz(iz[mask]/np.sqrt(2))
        ans[inv_mask] = np.exp(isigmak_cplx**2/2-isigmak_cplx*iz[inv_mask]) - \
            1/2*np.exp(-(t[inv_mask]/sigma)**2/2)*wofz(-iz[inv_mask]/np.sqrt(2))

    return ans.real, ans.imag

def dmp_osc_conv_cauchy_pair(t: Union[float, np.ndarray], fwhm: float,
k: float, T: float) -> Union[Tuple[float, float], Tuple[np.ndarray, np.ndarray]]:

    '''
    Compute damped oscillation convolved with normalized cauchy
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of cauchy distribution
      k: damping constant (inverse of life time)
      T: period of vibration 
      phase: phase factor

    Returns:
     Tuple of convolution of normalized cauchy distribution and 
     damped oscillation 
     :math:`(\\exp(-kt)cos(2\\pi t/T))` and :math:`(\\exp(-kt)sin(2\\pi t/T))`.
    '''

    gamma = fwhm/2; omega = 2*np.pi/T 
    z1 = (-k*t-omega*gamma) + complex(0,1)*(-k*gamma+omega*t)
    z2 = (-k*t+omega*gamma) + complex(0,-1)*(k*gamma+omega*t)
    ans1 = exp1x(z1)/(2*np.pi)+\
        np.exp(z1.real)*(-np.sin(z1.imag)+complex(0, 1)*np.cos(z1.imag))*np.heaviside(z1.imag, 1)
    ans2 = exp1x(z2)/(2*np.pi)
    return ans1.imag + ans2.imag, ans2.real - ans1.real

def dmp_osc_conv_gau(t: Union[float, np.ndarray], fwhm: float,
k: float, T: float, phase: float) -> Union[float, np.ndarray]:

    '''
    Compute damped oscillation convolved with normalized gaussian
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution
      k: damping constant (inverse of life time)
      T: period of vibration 
      phase: phase factor

    Returns:
     Convolution of normalized gaussian distribution and 
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    cosine, sine = exp_mod_gau_cplx(t, sigma, k, 2*np.pi/T)

    return cosine*np.cos(phase)-sine*np.sin(phase)

def dmp_osc_conv_cauchy(t: Union[float, np.ndarray], fwhm: float,
k: float, T: float, phase: float) -> Union[float, np.ndarray]:

    '''
    Compute damped oscillation convolved with normalized cauchy
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of cauchy distribution
      k: damping constant (inverse of life time)
      T: period of vibration 
      phase: phase factor

    Returns:
     Convolution of normalized cauchy distribution and 
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.
    '''

    cosine, sine = dmp_osc_conv_cauchy_pair(t, fwhm, k, T)

    return cosine*np.cos(phase)-sine*np.sin(phase)

def dmp_osc_conv_pvoigt(t: Union[float, np.ndarray], fwhm: float, eta: float,
k: float, T: float, phase: float) -> Union[float, np.ndarray]:

    '''
    Compute damped oscillation convolved with normalized pseudo
    voigt profile (i.e. linear combination of normalized gaussian and
    cauchy distribution)

    :math:`\\eta C(\\mathrm{fwhm}, t) + (1-\\eta)G(\\mathrm{fwhm}, t)`

    Args:
       t: time
       fwhm: uniform full width at half maximum parameter
       eta: mixing parameter
       k: damping constant (inverse of life time)
       T: period of vibration 
       phase: phase factor

    Returns:
     Convolution of normalized pseudo voigt profile and
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.
    '''

    u = dmp_osc_conv_gau(t, fwhm, k, T, phase)
    v = dmp_osc_conv_cauchy(t, fwhm, k, T, phase)

    return u + eta*(v-u)

def deriv_dmp_osc_conv_gau(t: Union[float, np.ndarray], fwhm: float,
k: float, T: float, phase: float) -> np.ndarray:

    '''
    Compute derivative of the convolution of damped oscillation and 
    normalized gaussian distribution

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution
      k: damping constant (inverse of life time)
      T: period of vibration 
      phase: phase factor

    Returns:
     Derivative of Convolution of normalized gaussian distribution and 
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.
    
    Note:

     * 1st column: df/dt
     * 2nd column: df/d(fwhm)
     * 3rd column: df/dk
     * 4th column: df/dT
     * 5th column: df/d(phase)
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2))); omega = 2*np.pi/T
    tmp1, tmp2 = exp_mod_gau_cplx(t, sigma, k, omega)
    freal = np.cos(phase)*tmp1 - np.sin(phase)*tmp2
    fimag = np.cos(phase)*tmp2 + np.sin(phase)*tmp1
    tmp3 = np.exp(-(t/sigma)**2/2)/np.sqrt(2*np.pi)
    greal = np.cos(phase)*tmp3; gimag = np.sin(phase)*tmp3

    if not isinstance(t, np.ndarray):
        grad = np.empty(5)
        grad[0] = greal/sigma - k*freal-omega*fimag
        grad[1] = (sigma*((k**2-omega**2)*freal + 2*k*omega*fimag) - 
        omega*gimag - (k+t/sigma**2)*greal)/(2*np.sqrt(2*np.log(2)))
        grad[2] = sigma**2*(k*freal+omega*fimag)-t*freal - sigma*greal
        grad[3] = -omega/T*(sigma**2*(k*fimag-omega*freal)-t*fimag - sigma*gimag)
        grad[4] = -fimag
    else:
        grad = np.empty((t.size, 5))
        grad[:, 0] = greal/sigma - k*freal-omega*fimag
        grad[:, 1] = (sigma*((k**2-omega**2)*freal + 2*k*omega*fimag) - 
        omega*gimag - (k+t/sigma**2)*greal)/(2*np.sqrt(2*np.log(2)))
        grad[:, 2] = sigma**2*(k*freal+omega*fimag)-t*freal - sigma*greal
        grad[:, 3] = -omega/T*(sigma**2*(k*fimag-omega*freal)-t*fimag - sigma*gimag)
        grad[:, 4] = -fimag
    
    return grad

def deriv_dmp_osc_conv_cauchy(t: Union[float, np.ndarray], fwhm: float,
k: float, T: float, phase: float) -> Union[float, np.ndarray]:

    '''
    Compute derivative of convolution of damped oscillation and normalized cauchy
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of cauchy distribution
      k: damping constant (inverse of life time)
      T: period of vibration 
      phase: phase factor

    Returns:
     Gradient of Convolution of normalized cauchy distribution and 
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.

    Note:

     * 1st column: df/dt
     * 2nd column: df/d(fwhm)
     * 3rd column: df/dk
     * 4th column: df/dT
     * 5th column: df/d(phase)
    '''

    gamma = fwhm/2; omega = 2*np.pi/T 
    z1 = (-k*t-omega*gamma) + complex(0,1)*(-k*gamma+omega*t)
    z2 = (-k*t+omega*gamma) + complex(0,-1)*(k*gamma+omega*t)

    f1 = complex(np.cos(phase), np.sin(phase))*exp1x(z1)/(2*np.pi)+np.exp(z1.real)*(-np.sin(z1.imag+phase)+complex(0,1)*np.cos(z1.imag+phase))*\
        np.heaviside(z1.imag, 1)
    f2 = complex(np.cos(phase), -np.sin(phase))*exp1x(z2)

    grad_z1 = f1 - complex(np.cos(phase), np.sin(phase))/(2*np.pi*z1)
    grad_z2 = (f2 - complex(np.cos(phase), -np.sin(phase))/z2)/(2*np.pi)
    sum = grad_z1 + grad_z2; diff = grad_z1 - grad_z2
    grad_t = -k*sum.imag + omega*diff.real
    grad_fwhm = -(omega*diff.imag+k*sum.real)/2
    grad_k = -(t*sum.imag+gamma*sum.real)
    grad_T = omega/T*(gamma*diff.imag-t*diff.real)
    grad_phase = f1.real-f2.real/(2*np.pi)

    if not isinstance(t, np.ndarray):
        grad = np.empty(5)
        grad[0] = grad_t
        grad[1] = grad_fwhm
        grad[2] = grad_k
        grad[3] = grad_T
        grad[4] = grad_phase
    else:
        grad = np.empty((t.size, 5))
        grad[:, 0] = grad_t
        grad[:, 1] = grad_fwhm
        grad[:, 2] = grad_k
        grad[:, 3] = grad_T
        grad[:, 4] = grad_phase
    
    return grad

def deriv_dmp_osc_sum_conv_gau(t: np.ndarray, fwhm: float,
k: np.ndarray, T: np.ndarray, phase: np.ndarray, c: np.ndarray) -> np.ndarray:

    '''
    Compute derivative of sum of damped oscillation function convolved with normalized gaussian
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of gaussian distribution
      k: rate constant (inverse of life time)
      T: period
      phase: phase factor
      c: coefficient

    Returns:
     Derivative of Convolution of normalized gaussian distribution and 
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.
    Note:
     * 1st column: df/dt
     * 2nd column: df/d(fwhm)
     * 2+i th column: df/dk_i :math:`(1 \\leq i \\leq {num}_{comp})`
     * 2+num_comp+i th column: df/dT_i :math:`(1 \\leq i \\leq {num}_{comp})`
     * :math:`2+2{num}_{comp}+i` th column: df/d(phase_i) :math:`(1 \\leq i \\leq {num}_{comp})`
    '''
    grad = np.zeros((t.size, 2+3*k.size))
    for i in range(k.size):
        grad_i = deriv_dmp_osc_conv_gau(t, fwhm, k[i], T[i], phase[i])
        grad[:, 0] = grad[:, 0] + c[i]*grad_i[:, 0]
        grad[:, 1] = grad[:, 1] + c[i]*grad_i[:, 1]
        grad[:, 2+i] = c[i]*grad_i[:, 2]
        grad[:, 2+k.size+i] = c[i]*grad_i[:, 3]
        grad[:, 2+2*k.size+i] = c[i]*grad_i[:, 4]

    return grad

def deriv_dmp_osc_sum_conv_cauchy(t: np.ndarray, fwhm: float,
k: np.ndarray, T: np.ndarray, phase: np.ndarray, c: np.ndarray) -> np.ndarray:

    '''
    Compute derivative of sum of damped oscillation function convolved with normalized cauchy
    distribution

    Args:
      t: time
      fwhm: full width at half maximum of cauchy distribution
      k: rate constant (inverse of life time)
      T: period
      phase: phase factor
      c: coefficient

    Returns:
     Derivative of Convolution of normalized cauchy distribution and 
     damped oscillation :math:`(\\exp(-kt)cos(2\\pi t/T+phase))`.
    Note:
     * 1st column: df/dt
     * 2nd column: df/d(fwhm)
     * 2+i th column: df/dk_i :math:`(1 \\leq i \\leq {num}_{comp})`
     * 2+num_comp+i th column: df/dT_i :math:`(1 \\leq i \\leq {num}_{comp})`
     * :math:`2+2{num}_{comp}+i` th column: df/d(phase_i) :math:`(1 \\leq i \\leq {num}_{comp})`
    '''
    grad = np.zeros((t.size, 2+3*k.size))
    for i in range(k.size):
        grad_i = deriv_dmp_osc_conv_cauchy(t, fwhm, k[i], T[i], phase[i])
        grad[:, 0] = grad[:, 0] + c[i]*grad_i[:, 0]
        grad[:, 1] = grad[:, 1] + c[i]*grad_i[:, 1]
        grad[:, 2+i] = c[i]*grad_i[:, 2]
        grad[:, 2+k.size+i] = c[i]*grad_i[:, 3]
        grad[:, 2+2*k.size+i] = c[i]*grad_i[:, 4]

    return grad







import argparse
import numpy as np
import matplotlib.pyplot as plt
from ..mathfun import voigt_thy

description = '''
calc_broad: Evaluates voigt broadened theoritical calc spectrum
'''

policy_help = '''
Policy to match discrepency between experimental data and theoretical spectrum.
 'shift': constant shift peak position
 'scale': constant scale peak position
 'both': shift and scale peak position
'''

def calc_broad():

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('peak',
                        help='filename for calculated line shape spectrum')
    parser.add_argument('e_min', type=float,
                        help='minimum energy')
    parser.add_argument('e_max', type=float,
                        help='maximum energy')
    parser.add_argument('e_step', type=float,
    help='energy step')
    parser.add_argument('A', type=float,
                        help='scale factor')
    parser.add_argument('fwhm_G', type=float,
                        help='Full Width at Half Maximum of gaussian shape')
    parser.add_argument('fwhm_L', type=float,
                        help='Full Width at Half Maximum of lorenzian shape')
    parser.add_argument('--shift_factor', type=float,
    help='parameter to shift peak position of thoretical spectrum')
    parser.add_argument('--scale_factor', type=float,
    help='paramter to scale peak position of theoretical spectrum')
    parser.add_argument('-o', '--out', help='prefix for output files')
    parser.add_argument('--policy', choices=['shift', 'scale', 'both'], type=str,
    help=policy_help)
    args = parser.parse_args()

    peak = np.genfromtxt(args.peak)[:,:2]
    if args.out is None:
        out = args.prefix
    else:
        out = args.out
    e_min = args.e_min
    e_max = args.e_max
    e_step = args.e_step
    A = args.A
    fwhm_G = args.fwhm_G
    fwhm_L = args.fwhm_L
    if args.policy is None:
        policy = 'shift'
        peak_factor = 0
        args.shift_factor = 0
    else:
        policy = args.policy
    
    if policy in ['shift', 'both'] and args.shift_factor is None:
        raise Exception(f"Your policy is {args.policy}, please set shift_factor parameter.")
    if policy in ['scale', 'both'] and args.scale_factor is None:
        raise Exception(f"Your policy is {args.policy}, please set scale_factor parameter.")

    if args.policy == 'shift':
        peak_factor = args.shift_factor
    elif args.policy == 'scale':
        peak_factor = args.scale_factor
    else:
        peak_factor = np.array([args.shift_factor, args.scale_factor])

    e = np.linspace(e_min, e_max, int((e_max-e_min)/e_step)+1)

    broadened_thy = A*voigt_thy(e, peak, fwhm_G, fwhm_L, peak_factor, policy)

    rescaled_stk = peak.copy()
    if policy == 'scale':
        rescaled_stk[:, 0] = rescaled_stk[:, 0]*peak_factor
    elif policy == 'shift':
        rescaled_stk[:, 0] = rescaled_stk[:, 0] - peak_factor
    else:
        rescaled_stk[:, 0] = peak_factor[1]*rescaled_stk[:, 0] - peak_factor[0]
    rescaled_stk[:, 1] = A*rescaled_stk[:, 1]/np.sum(rescaled_stk[:, 1])
    spec_thy = np.vstack((e, broadened_thy)).T

    np.savetxt(f'{out}_thy_stk.txt', rescaled_stk, fmt=['%.8e', '%.8e'], header='energy \t abs')
    np.savetxt(f'{out}_thy.txt', spec_thy, fmt=['%.8e', '%.8e'], header='energy \t abs')

    plt.plot(e, broadened_thy)
    plt.title("voigt broadened theoretical spectrum")
    plt.show()

    return



'''
_transient_dmp_osc:
submodule for fitting time delay scan with the
convolution of sum of damped oscillation and instrumental response function 

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''
from typing import Optional, Union, Sequence, Tuple
import numpy as np
from ..mathfun.irf import calc_eta, calc_fwhm
from .transient_result import TransientResult
from scipy.optimize import basinhopping
from scipy.optimize import least_squares
from ..mathfun.A_matrix import make_A_matrix_dmp_osc, fact_anal_A
from ..res.parm_bound import set_bound_t0, set_bound_tau
from ..res.res_osc import residual_dmp_osc, res_grad_dmp_osc

def fit_transient_dmp_osc(irf: str, fwhm_init: Union[float, np.ndarray], 
                          t0_init: np.ndarray, tau_init: np.ndarray, period_init: np.ndarray,
                          do_glb: Optional[bool] = False, 
                          method_lsq: Optional[str] = 'trf',
                          kwargs_glb: Optional[dict] = None, 
                          kwargs_lsq: Optional[dict] = None,
                          bound_fwhm: Optional[Sequence[Tuple[float, float]]] = None, 
                          bound_t0: Optional[Sequence[Tuple[float, float]]] = None, 
                          bound_tau: Optional[Sequence[Tuple[float, float]]] = None,
                          bound_period: Optional[Sequence[Tuple[float, float]]] = None,
                          name_of_dset: Optional[Sequence[str]] = None,
                          t: Optional[Sequence[np.ndarray]] = None, 
                          intensity: Optional[Sequence[np.ndarray]] = None,
                          eps: Optional[Sequence[np.ndarray]] = None) -> TransientResult:
                      
      '''
      driver routine for fitting multiple data set of time delay scan data with
      sum of the convolution of dmped oscillation and instrumental response function.

      It separates linear and non-linear part of the optimization problem to solve non linear least sequare
      optimization problem efficiently.

      Moreover, this driver uses two step method to search best parameter, its covariance and
      estimated parameter error.

      Step 1. (basinhopping)
      Use global optimization to find rough global minimum of our objective function.
      In this stage, it use analytic gradient for scalar residual function.

      Step 2. (method_lsq)
      Use least squares optimization algorithm to refine global minimum of objective function and approximate covariance matrix.
      Because of linear and non-linear seperation scheme, the analytic jacobian for vector residual function is hard to obtain.
      Thus, in this stage, it uses numerical jacobian.               
      
      Args:
       irf ({'g', 'c', 'pv'}): shape of instrumental response function
       
        'g': gaussian shape
        
        'c': cauchy shape
        
        'pv': pseudo voigt shape (kind: 2)
       fwhm_init (float or np.ndarray): initial full width at half maximum for instrumental response function
       
        * if irf in ['g', 'c'] then fwhm_init is float
        * if irf == 'pv' then fwhm_init is the `numpy.ndarray` with [fwhm_G, fwhm_L]

       t0_init (np.ndarray): time zeros for each scan
       tau_init (np.ndarray): lifetime constant for each damped oscillation component
       period_init (np.ndarray): period of each oscillation component
       do_glb (bool): Whether or not use global optimization algorithm. If True then basinhopping algorithm is used.
       method_lsq ({'trf', 'dogbox', 'lm'}): method of local optimization for least_squares
                                             minimization (refinement of global optimization solution)
       kwargs_glb: keyward arguments for global optimization solver
       kwargs_lsq: keyward arguments for least square optimization solver
       bound_fwhm (sequence of tuple): boundary for irf parameter. If upper and lower bound are same, 
        driver assumes that the parameter is fixed during optimization. If `bound_fwhm` is `None`, 
        the upper and lower bound are given as `(fwhm_init/2, 2*fwhm_init)`.
       bound_t0 (sequence of tuple): boundary for time zero parameter. 
        If `bound_t0` is `None`, the upper and lower bound are given as `(t0-2*fwhm_init, t0+2*fwhm_init)`.
       bound_tau (sequence of tuple): boundary for lifetime constant for damped oscillation component, 
        if `bound_tau` is `None`, the upper and lower bound are given by `set_bound_tau`.
       bound_period (sequence of tuple): boundary for period of damped oscillation component, 
        if `bound_period` is `None`, the upper and lower bound are given by `set_bound_tau`.
       name_of_dset (sequence of str): name of each dataset
       t (sequence of np.narray): time scan range for each datasets
       intensity (sequence of np.ndarray): sequence of intensity pf datasets for time delay scan (it should not contain time scan range)
       eps (sequence of np.ndarray): sequence of estimated errors of each dataset

       Returns:
        TransientResult class object
      '''
      num_comp = tau_init.size

      num_irf = 1*(irf in ['g', 'c'])+2*(irf == 'pv')
      num_param = num_irf+t0_init.size+2*num_comp
      param = np.empty(num_param, dtype=float)
      fix_param_idx = np.empty(num_param, dtype=bool)

      param[:num_irf] = fwhm_init
      param[num_irf:num_irf+t0_init.size] = t0_init
      param[num_irf+t0_init.size:num_irf+t0_init.size+num_comp] = tau_init
      param[num_irf+t0_init.size+num_comp:num_irf+t0_init.size+2*num_comp] = period_init
      bound = num_param*[None]

      if bound_fwhm is None:
            for i in range(num_irf):
                  bound[i] = (param[i]/2, 2*param[i])
      else:
            bound[:num_irf] = bound_fwhm
                      
      if bound_t0 is None:
            for i in range(t0_init.size):
                  bound[i+num_irf] = set_bound_t0(t0_init[i], fwhm_init)
      else:
            bound[num_irf:num_irf+t0_init.size] = bound_t0
      
      if bound_tau is None:
            for i in range(num_comp):
                  bound[i+num_irf+t0_init.size] = set_bound_tau(tau_init[i], fwhm_init)
      else:
            bound[num_irf+t0_init.size:num_irf+t0_init.size+num_comp] = bound_tau
      
      if bound_period is None:
            for i in range(num_comp):
                  bound[i+num_irf+t0_init.size+num_comp] = set_bound_tau(period_init[i], fwhm_init)
      else:
            bound[num_irf+t0_init.size+num_comp:num_irf+t0_init.size+2*num_comp] = bound_period

      for i in range(num_param):
            fix_param_idx[i] = (bound[i][0] == bound[i][1])

      if do_glb:
            go_args = (num_comp, irf, fix_param_idx, t, intensity, eps)
            min_go_kwargs = {'args': go_args, 'jac': True, 'bounds': bound}
            if kwargs_glb is not None:
                  minimizer_kwargs = kwargs_glb.pop('minimizer_kwargs', None)
                  if minimizer_kwargs is None:
                        kwargs_glb['minimizer_kwargs'] = min_go_kwargs
                  else:
                        minimizer_kwargs['args'] = min_go_kwargs['args']
                        minimizer_kwargs['jac'] = min_go_kwargs['jac']
                        minimizer_kwargs['bounds'] = min_go_kwargs['bounds']
                        kwargs_glb['minimizer_kwargs'] = minimizer_kwargs
            else:
                  kwargs_glb = {'minimizer_kwargs' : min_go_kwargs}
            res_go = basinhopping(res_grad_dmp_osc, param, **kwargs_glb)
      else:
            res_go = dict()
            res_go['x'] = param
            res_go['message'] = None
            res_go['nfev'] = 0

      param_gopt = res_go['x']
      args_lsq = (num_comp, irf, t, intensity, eps)

      if kwargs_lsq is not None:
            _ = kwargs_lsq.pop('args', None)
            _ = kwargs_lsq.pop('kwargs', None)
            kwargs_lsq['args'] = args_lsq
      else:
            kwargs_lsq = {'args' : args_lsq}

      bound_tuple = (num_param*[None], num_param*[None])
      for i in range(num_param):
            bound_tuple[0][i] = bound[i][0]
            bound_tuple[1][i] = bound[i][1]
            if bound[i][0] == bound[i][1]:
                  if bound[i][0] > 0:
                        bound_tuple[1][i] = bound[i][1]*(1+1e-8)+1e-16
                  else:
                        bound_tuple[1][i] = bound[i][1]*(1-1e-8)+1e-16

      # jacobian of vector residual function is inaccurate
      res_lsq = least_squares(residual_dmp_osc, param_gopt, method=method_lsq, bounds=bound_tuple, **kwargs_lsq)
      param_opt = res_lsq['x']

      fwhm_opt = param_opt[:num_irf]
      tau_opt = param_opt[num_irf+t0_init.size:num_irf+t0_init.size+num_comp]
      period_opt = param_opt[num_irf+t0_init.size+num_comp:num_irf+t0_init.size+2*num_comp]
      
      fit = np.empty(len(t), dtype=object); res = np.empty(len(t), dtype=object)
      
      num_tot_scan = 0
      for i in range(len(t)):
            num_tot_scan = num_tot_scan + intensity[i].shape[1]
            fit[i] = np.empty(intensity[i].shape)
            res[i] = np.empty(intensity[i].shape)

    # Calc individual chi2
      chi = res_lsq['fun']
      num_param_tot = 2*num_tot_scan*num_comp+num_param-np.sum(fix_param_idx)
      chi2 = 2*res_lsq['cost']
      red_chi2 = chi2/(chi.size-num_param_tot)
      
      start = 0; end = 0; 
      chi2_ind = np.empty(len(t), dtype=object); red_chi2_ind = np.empty(len(t), dtype=object)
      num_param_ind = 4*num_comp+2+1*(irf == 'pv')

      for i in range(len(t)):
            step = intensity[i].shape[0]
            chi2_ind_aux = np.empty(intensity[i].shape[1], dtype=float)
            for j in range(intensity[i].shape[1]):
                  end = start + step
                  chi2_ind_aux[j] = np.sum(chi[start:end]**2)
                  start = end
            chi2_ind[i] = chi2_ind_aux
            red_chi2_ind[i] = chi2_ind[i]/(intensity[i].shape[0]-num_param_ind)

      param_name = np.empty(param_opt.size, dtype=object)
      c = np.empty(len(t), dtype=object)
      phase = np.empty(len(t), dtype=object)
      t0_idx = num_irf

      if irf == 'g':
            fwhm_pv = fwhm_opt[0]
            eta = 0
            param_name[0] = 'fwhm_G'
      elif irf == 'c':
            fwhm_pv = fwhm_opt[0]
            eta = 1
            param_name[0] = 'fwhm_L'
      else:
            fwhm_pv = calc_fwhm(fwhm_opt[0], fwhm_opt[1])
            eta = calc_eta(fwhm_opt[0], fwhm_opt[1])
            param_name[0] = 'fwhm_G'
            param_name[1] = 'fwhm_L'

      for i in range(len(t)):
            c[i] = np.empty((num_comp, intensity[i].shape[1]))
            phase[i] = np.empty((num_comp, intensity[i].shape[1]))
            
            for j in range(intensity[i].shape[1]):
                  A = make_A_matrix_dmp_osc(t[i]-param_opt[t0_idx], fwhm_pv, tau_opt, period_opt, irf, eta)
                  tmp = fact_anal_A(A, intensity[i][:, j], eps[i][:, j])
                  c[i][:, j] = np.sqrt(tmp[:num_comp]**2+tmp[num_comp:])
                  phase[i][:, j] = -np.arctan2(tmp[num_comp:], tmp[:num_comp]) 
                  fit[i][:, j] = tmp @ A
                  param_name[t0_idx] = f't_0_{i+1}_{j+1}'
                  t0_idx = t0_idx + 1
            
            res[i] = intensity[i] - fit[i]
      
      for i in range(num_comp):
            param_name[num_irf+t0_init.size+i] = f'tau_{i+1}'
            param_name[num_irf+t0_init.size+num_comp+i] = f'period_{i+1}'
      
      jac = res_lsq['jac']
      hes = jac.T @ jac
      cov = np.zeros_like(hes)
      n_free_param = np.sum(~fix_param_idx)
      mask_2d = np.einsum('i,j->ij', ~fix_param_idx, ~fix_param_idx)
      cov[mask_2d] = np.linalg.inv(hes[mask_2d].reshape((n_free_param, n_free_param))).flatten()
      cov_scaled = red_chi2*cov
      param_eps = np.sqrt(np.diag(cov_scaled))
      corr = cov_scaled.copy()
      weight = np.einsum('i,j->ij', param_eps, param_eps)
      corr[mask_2d] = corr[mask_2d]/weight[mask_2d]

      result = TransientResult()
      # save experimental fitting data
      if name_of_dset is None:
            name_of_dset = np.empty(len(t), dtype=object)
            for i in range(len(t)):
                  name_of_dset[i] = f'dataset_{i+1}'
                  
      result['name_of_dset'] = name_of_dset; result['t'] = t
      result['intensity'] = intensity; result['eps'] = eps

      result['model'] = 'dmp_osc'
      result['fit'] = fit; result['res'] = res; result['irf'] = irf
      result['fwhm'] = fwhm_pv; result['eta'] = eta
      
      result['param_name'] = param_name; result['x'] = param_opt
      result['bounds'] = bound; result['base'] = False; result['c'] = c
      result['phase'] = phase
      result['chi2'] = chi2; result['chi2_ind'] = chi2_ind
      result['aic'] = chi.size*np.log(chi2/chi.size)+2*num_param_tot
      result['bic'] = chi.size*np.log(chi2/chi.size)+num_param_tot*np.log(chi.size)
      result['red_chi2'] = red_chi2; result['red_chi2_ind'] = red_chi2_ind
      result['nfev'] = res_go['nfev'] + res_lsq['nfev']
      result['n_param'] = num_param_tot; result['n_param_ind'] = num_param_ind
      result['num_pts'] = chi.size; result['jac'] = jac
      result['cov'] = cov; result['corr'] = corr; result['cov_scaled'] = cov_scaled
      result['x_eps'] = param_eps
      result['method_lsq'] = method_lsq
      result['message_lsq'] = res_lsq['message']
      result['success_lsq'] = res_lsq['success']

      if result['success_lsq']:
            result['status'] = 0
      else:
            result['status'] = -1
      
      if do_glb:
            result['method_glb'] = 'basinhopping'
            result['message_glb'] = res_go['message'][0]  
      else:
            result['method_glb'] = None
            result['message_glb'] = None

      result['n_osc'] = tau_init.size
      result['n_decay'] = 0

      return result


'''
A_matrix:
submodule for evaluation of A_matrix

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''
from typing import Optional
import numpy as np
from scipy.linalg import lstsq
from .exp_conv_irf import exp_conv_gau, exp_conv_cauchy
from .exp_conv_irf import exp_mod_gau_cplx, dmp_osc_conv_cauchy_pair


def make_A_matrix(t: np.ndarray, k: np.ndarray) -> np.ndarray:

    A = np.empty((k.size, t.size))
    for i in range(k.size):
        A[i, :] = np.exp(-k[i]*t)
    A = A*np.heaviside(t, 1)
    A[np.isnan(A)] = 0

    return A


def make_A_matrix_gau(t: np.ndarray, fwhm: float,
                      k: np.ndarray) -> np.ndarray:

    A = np.empty((k.size, t.size))
    for i in range(k.size):
        A[i, :] = exp_conv_gau(t, fwhm, k[i])

    return A

def make_A_matrix_cauchy(t: np.ndarray, fwhm: float,
                         k: np.ndarray) -> np.ndarray:

    A = np.empty((k.size, t.size))
    for i in range(k.size):
        A[i, :] = exp_conv_cauchy(t, fwhm, k[i])

    return A


def make_A_matrix_pvoigt(t: np.ndarray,
                         fwhm: float,
                         eta: float,
                         k: np.ndarray) -> np.ndarray:
    
    u = make_A_matrix_gau(t, fwhm, k)
    v = make_A_matrix_cauchy(t, fwhm, k)

    return u + eta*(v-u)

def make_A_matrix_exp(t: np.ndarray,
                      fwhm: float,
                      tau: np.ndarray,
                      base: Optional[bool] = True,
                      irf: Optional[str] = 'g',
                      eta: Optional[float] = None
                      ) -> np.ndarray:
    
    if base:
        k = np.empty(tau.size+1)
        k[-1] = 0; k[:-1] = 1/tau
    else:
        k = 1/tau

    if irf == 'g':
        A = make_A_matrix_gau(t, fwhm, k)
    elif irf == 'c':
        A = make_A_matrix_cauchy(t, fwhm, k)
    else:
        A = make_A_matrix_pvoigt(t, fwhm, eta, k)

    return A

def make_A_matrix_gau_osc(t: np.ndarray, fwhm: float,
k: np.ndarray, T: np.ndarray) -> np.ndarray:

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    A = np.empty((2*k.size, t.size))
    for i in range(k.size):
        A[i, :], A[i+k.size, :] = \
            exp_mod_gau_cplx(t, sigma, k[i], 2*np.pi/T[i])
        # cosine, sine
    return A

def make_A_matrix_cauchy_osc(t: np.ndarray, fwhm: float,
k: np.ndarray, T: np.ndarray) -> np.ndarray:

    A = np.empty((2*k.size, t.size))
    for i in range(k.size):
        A[i, :], A[i+k.size, :] = \
            dmp_osc_conv_cauchy_pair(t, fwhm, k[i], T[i])
    return A

def make_A_matrix_pvoigt_osc(t: np.ndarray, fwhm: float, eta: float,
                             k: np.ndarray, 
                             T: np.ndarray) -> np.ndarray:
    
    u = make_A_matrix_gau_osc(t, fwhm, k, T)
    v = make_A_matrix_cauchy_osc(t, fwhm, k, T)

    return u + eta*(v-u)

def make_A_matrix_dmp_osc(t: np.ndarray, fwhm: float,
                      tau: np.ndarray,
                      T: np.ndarray,
                      irf: Optional[str] = 'g',
                      eta: Optional[float] = None
                      ) -> np.ndarray:


    if irf == 'g':
        A = make_A_matrix_gau_osc(t, fwhm, 1/tau, T)
    elif irf == 'c':
        A = make_A_matrix_cauchy_osc(t, fwhm, 1/tau, T)
    elif irf == 'pv':
        A = make_A_matrix_pvoigt_osc(t, fwhm, eta, 1/tau, T)

    return A

def fact_anal_A(A: np.ndarray, 
                intensity: Optional[np.ndarray] = None, 
                eps: Optional[np.ndarray] = None) -> np.ndarray:

    if eps is None:
        eps = np.ones_like(intensity)
    
    B = np.einsum('j,ij->ij', 1/eps, A)
    y = intensity/eps
    
    c, _, _, _ = lstsq(B.T, y, cond=1e-2)
    return c
